/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     coarsen.c                                                      */
/*                                                                          */
/*    !!!  includes ../Common/coarsen_common.c  !!!                         */
/*                                                                          */
/* description:  coarsening of 1 dim. hierarchical meshes;                  */
/*               at the end of file Common/coarsen_common.c is included;    */
/*               file contains all routines depending on DIM == 1;          */
/*               routines for both dimensions in ../Common/coarsen_common.c */
/*               are included at the end of the file                        */
/*                                                                          */
/* global variables that have to be defined for Common/coarsen_common.c:    */
/*  static MESH           *coarse_mesh;                                     */
/*  static int            do_more;                                          */
/*  static int            call_coarse_restrict;                             */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/*--------------------------------------------------------------------------*/
/*  global variables used for coarsening                                    */
/*--------------------------------------------------------------------------*/

static int   call_coarse_restrict;
static void  coarse_restrict(RC_LIST_EL *, int);

/*--------------------------------------------------------------------------*/
/*  coarse_element:  coarses a single element of the coarsening patch; dofs */
/*  in the interior of the element are removed; dofs for higher order       */
/*  at the boundary or the coarsening patch still belong to                 */
/*  the parent. Do not remove them form the mesh!!!                         */
/*                                                                          */
/*  if NEIGH_IN_EL neighbour information to elements outside the patch is   */
/*  set on the parent;                                                      */
/*--------------------------------------------------------------------------*/

static int coarse_recursive(MESH *mesh, EL *parent)
{
  FUNCNAME("coarse_recursive");
  int  mark;

  INFO(0,2)("\n");
  if (!parent->child[0])  /*---  leaf element, clean element marker      ---*/
  {
    mark = parent->mark;
    parent->mark = 0;
    return(mark);
  }
  else                    /*---  no leaf element, first coarse children  ---*/
  {
    EL       *child[2];

    int  mark0 = coarse_recursive(mesh, parent->child[0]);
    int  mark1 = coarse_recursive(mesh, parent->child[1]);

    mark = MAX(mark0, mark1);

    child[0] = parent->child[0];   child[1] = parent->child[1]; 

    if (mark >= 0)         /*--- element must not be coarsend            ---*/
    {
      parent->mark = 0;
      if (child[0])
	child[0]->mark = child[1]->mark = 0;

      return(0);
    }

/*--------------------------------------------------------------------------*/
/*--- and now coarsen child[0] and child[1] into parent                  ---*/
/*--------------------------------------------------------------------------*/
  
#if NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  adjust neighbour connectivity  at the boundary of the coarsening patch  */
/*--------------------------------------------------------------------------*/
    {
      EL     *neigh;
      int    opp_v, i_child;

      for (i_child = 0; i_child < 2; i_child++)
      {
	if (neigh = (el->neigh[1-i_child] = child[i_child]->neigh[1-i_child]))
	{
	  opp_v = child[i_child]->opp_vertex[1-i_child];
	  el->opp_vertex[1-i_child] = opp_v;
	  neigh->neigh[opp_v] = el;
	  neigh->opp_vertex[opp_v] = 1-i_child;
	}
      }
    }
#endif

/*--------------------------------------------------------------------------*/
/*--- hand DOFs from children to parent, and add DOF at center           ---*/
/*--------------------------------------------------------------------------*/

    if (mesh->n_dof[VERTEX] && !mesh->preserve_coarse_dofs)
    {
      int node = mesh->node[VERTEX];
      parent->dof[node+0] = child[0]->dof[node+0];
      parent->dof[node+1] = child[1]->dof[node+1];
    }

    if (mesh->n_dof[CENTER] && !mesh->preserve_coarse_dofs)
    {
      parent->dof[mesh->node[CENTER]] = get_dof(mesh, CENTER);
    }

/*--------------------------------------------------------------------------*/
/*  restrict dof vectors to the parents on the patch                        */
/*--------------------------------------------------------------------------*/

    if (call_coarse_restrict)
    {
      RC_LIST_EL  coarse_list[1] = {{nil, 0, 0}};
      coarse_list->el = parent;
      coarse_restrict(coarse_list, 1);
    }

/*--------------------------------------------------------------------------*/
/*--- remove all DOFs of children that are not used anymore              ---*/
/*--------------------------------------------------------------------------*/

    if (mesh->n_dof[VERTEX])          /*---  midpoint of parent          ---*/
    {
      free_dof(child[1]->dof[mesh->node[VERTEX]], mesh, VERTEX);
    }

    if (mesh->n_dof[CENTER])          /*--- center of the children       ---*/
    {
      free_dof(child[0]->dof[mesh->node[CENTER]], mesh, CENTER);
      free_dof(child[1]->dof[mesh->node[CENTER]], mesh, CENTER);
    }

    parent->child[0] = nil;
    parent->child[1] = get_leaf_data(mesh);

    if (parent->child[1]  &&  mesh->leaf_data_info->coarsen_leaf_data)
      (*mesh->leaf_data_info->coarsen_leaf_data)(parent, child);

    free_element(child[0], mesh);
    free_element(child[1], mesh);


    mesh->n_elements--;
    mesh->n_hier_elements -= 2;
    mesh->n_vertices--;

    mark++;
    parent->mark = MIN(mark,0);

    return(parent->mark);
  }

  return(0);  /*--- statement never reached                              ---*/
}

#include "coarsen_common.c"
