/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     graphXO.c                                                      */
/*                                                                          */
/*                                                                          */
/* description:  simple graphical routines in 1d                            */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

const GRAPH_RGBCOLOR rgb_black   = {0.0,0.0,0.0};
const GRAPH_RGBCOLOR rgb_white   = {1.0,1.0,1.0};
const GRAPH_RGBCOLOR rgb_red     = {1.0,0.0,0.0};
const GRAPH_RGBCOLOR rgb_green   = {0.0,1.0,0.0};
const GRAPH_RGBCOLOR rgb_blue    = {0.0,0.0,1.0};
const GRAPH_RGBCOLOR rgb_yellow  = {1.0,1.0,0.0};
const GRAPH_RGBCOLOR rgb_magenta = {1.0,0.0,1.0};
const GRAPH_RGBCOLOR rgb_cyan    = {0.0,1.0,1.0};
const GRAPH_RGBCOLOR rgb_grey50  = {0.5,0.5,0.5};

const GRAPH_RGBCOLOR rgb_albert  = {0.433,0.781,0.969};
const GRAPH_RGBCOLOR rgb_alberta = {0.000,0.391,0.605};

#include <GL/glx.h>
#include <GL/gl.h>
#include <unistd.h>

struct ogl_window
{
  Display      *dpy;
  Window       win;
  GLXContext   cx;
  unsigned int width, height;
};
typedef struct ogl_window OGL_WINDOW;
static Bool       opengl_available = True;

/****************************************************************************/

static Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{
  return (e->type == MapNotify) && (e->xmap.window == (Window)arg);
}

/****************************************************************************/

static Bool OGL_set_std_window(OGL_WINDOW *ogl_win)
{
  if (ogl_win) 
    return(glXMakeCurrent(ogl_win->dpy, ogl_win->win, ogl_win->cx));
  else
    return(false);
}

static OGL_WINDOW *OGL_create_window(const char *name, const char *geometry)
{
  FUNCNAME("OGL_create_window");
  static Display     *dpy = nil;
  static XVisualInfo *vi  = nil;
  static Colormap    cmap;
  static GLXContext  cx;
  static char        *default_geometry = "=300x300+0+0";
  static int         attributeList[] = { GLX_RGBA, None };
  XSetWindowAttributes swa;
  OGL_WINDOW         *ogl_win;
  Window             win, root;
  XSizeHints         size_hints;
  XEvent             event;
  unsigned int       bwidth, depth;
  int                status, x, y;
  
  if (!dpy) /* get a connection */
  {
    if (!(dpy = XOpenDisplay(nil))) 
    {
      MSG("cannot connect to display.\n");
      return(nil);
    }

    opengl_available = glXQueryExtension(dpy, NULL, NULL);
    if (!opengl_available)
    {
      MSG("no OpenGL available...\n");
      return(nil);
    }
  }

  
  if (!vi) /* get an appropriate visual */
  {
    vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);
    if (!vi) 
    {
      MSG("no visual\n");
      return(nil);
    }
    cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen),
			   vi->visual, AllocNone);
    if (!cmap) 
    {
      MSG("no color map\n");
      return(nil);
    }
  }
  

  size_hints.flags = PResizeInc | PMinSize ;
  size_hints.width_inc  =  1;
  size_hints.height_inc =  1;
  size_hints.min_width  = 10;
  size_hints.min_height = 10;

  if (geometry) 
  {
    status = XGeometry(dpy, vi->screen, geometry, default_geometry,
		       2, 1, 1, 0, 0, 
		       &(size_hints.x), &(size_hints.y),
		       &(size_hints.width), &(size_hints.height));
    if (status & (XValue|YValue)) {
      size_hints.flags |= USPosition;
      size_hints.flags &= ~PPosition;
    }
    if (status & (WidthValue|HeightValue)) 
    {
      size_hints.flags |= USSize;
      size_hints.flags &= ~PSize;
    }
  }

  if (size_hints.flags & USSize) 
  {
    size_hints.flags |= PAspect;
    size_hints.min_aspect.x = size_hints.width;
    size_hints.min_aspect.y = size_hints.height;
    size_hints.max_aspect.x = size_hints.width;
    size_hints.max_aspect.y = size_hints.height;
  }

  /* create a window */
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = StructureNotifyMask;
  win = XCreateWindow(dpy, RootWindow(dpy, vi->screen),
		      size_hints.x, size_hints.y,
		      size_hints.width, size_hints.height,
		      0, vi->depth, InputOutput, vi->visual,
		      CWBorderPixel|CWColormap|CWEventMask, &swa);
  if (!win) 
  {
    MSG("no window\n");
    return(nil);
  }

  XSetStandardProperties(dpy, win, name, name, None, NULL, 0, &size_hints);

  XMapWindow(dpy, win);
  XIfEvent(dpy, &event, WaitForNotify, (char*)win);


  /* create a GLX context */
  cx = glXCreateContext(dpy, vi, 0, GL_TRUE);
  if (!cx) 
  {
    MSG("no GLX context\n");
    return(nil);
  }

  /* connect the context to the window */
  glXMakeCurrent(dpy, win, cx);

  /* clear the buffer */
  glClearColor(0,0,0,1);
  glClear(GL_COLOR_BUFFER_BIT);
  glFlush();

  ogl_win = MEM_ALLOC(1, OGL_WINDOW);
  ogl_win->dpy = dpy;
  ogl_win->win = win;
  ogl_win->cx  = cx;
  XGetGeometry(dpy, win, &root, &x, &y,
	       &ogl_win->width, &ogl_win->height,
	       &bwidth, &depth);
  
  return(ogl_win);
}

/****************************************************************************/

static void OGL_destroy_window(OGL_WINDOW *ogl_win)
{
  if (!ogl_win) return;

  glXDestroyContext(ogl_win->dpy, ogl_win->cx);

  XDestroyWindow(ogl_win->dpy, ogl_win->win);
  XFlush(ogl_win->dpy);

  MEM_FREE(ogl_win, 1, OGL_WINDOW);

  return;
}

/****************************************************************************/

static void OGL_clear_window(OGL_WINDOW *ogl_win, const GRAPH_RGBCOLOR c)
{
  if (!ogl_win) return;

  if (OGL_set_std_window(ogl_win))
  {
    Window       root;
    unsigned int w, h, depth, bwidth;
    int          x, y;

/*--- has the window size changed?                                       ---*/
    XGetGeometry(ogl_win->dpy, ogl_win->win, &root, &x, &y, &w, &h, &bwidth,
		 &depth);
 
    if (w != ogl_win->width || h != ogl_win->height)
    {
      ogl_win->width  = w;
      ogl_win->height = h;
      XResizeWindow(ogl_win->dpy, ogl_win->win, w, h);
    }

    glXMakeCurrent(ogl_win->dpy, ogl_win->win, ogl_win->cx);

    glViewport(0, 0, w, h);
    if (c)
      glClearColor(c[0], c[1], c[2], 1.0);
    else
      glClearColor(1.0, 1.0, 1.0, 1.0);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glFlush();
  }
  return;
}

void graph_close_window(GRAPH_WINDOW win)
{
  OGL_destroy_window((OGL_WINDOW *)win);
  return;
}

/*--------------------------------------------------------------------------*/
/*---  get min and max coordinates                                       ---*/
/*--------------------------------------------------------------------------*/
static float xmin[2], xmax[2];

static void xminmax_fct(const EL_INFO *el_info)
{
  FUNCNAME("xminmax_fct");
  int        i, j;
  PARAMETRIC *parametric = el_info->mesh->parametric;
  
  if (parametric && parametric->init_element(el_info, parametric)) 
  {
    const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};
    REAL_D     world[N_VERTICES];

    parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);

    for (i = 0; i < N_VERTICES; i++) 
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xmin[j] = MIN(xmin[j], world[i][j]);
	xmax[j] = MAX(xmax[j], world[i][j]);
      }
    }
  }
  else 
  {
    TEST_FLAG(FILL_COORDS, el_info);

    for (i = 0; i < N_VERTICES; i++)
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xmin[j] = MIN(xmin[j], el_info->coord[i][j]);
	xmax[j] = MAX(xmax[j], el_info->coord[i][j]);
      }    
    }
  }

  return;
}

GRAPH_WINDOW graph_open_window(const char *title, const char *geom, 
			       REAL *world, MESH *mesh)
{
  FUNCNAME("graph_open_window");
  OGL_WINDOW *winO = nil;
  int        i;
  char       geometry[16];
  float      xdiam[2];


  if (DIM!=DIM_OF_WORLD) {
    WARNING("Sorry, graphic not implemented for DIM!=DIM_OF_WORLD!\n");
    return(nil);
  }
    
  if (world)
  {
    xmin[0] = world[0];
    xmax[0] = world[1];
    xmin[1] = world[2];
    xmax[1] = world[3];
    xdiam[0] = MAX(xmax[0] - xmin[0], 1.E-10);
    xdiam[1] = MAX(xmax[1] - xmin[1], 1.E-10);
  }
  else if (mesh)
  {
    for (i = 0; i < DIM_OF_WORLD; i++)
      xmax[i] = -(xmin[i] = 1.0E10);
    mesh_traverse(mesh, -1, CALL_LEAF_EL|FILL_COORDS, xminmax_fct);

    xmin[1] = -1.0;
    xmax[1] = 1.0;

    for (i = 0; i < DIM_OF_WORLD; i++)
    {
      xdiam[i] = MAX(xmax[i] - xmin[i], 1.E-10);
      xmin[i] -= 0.1*xdiam[i];
      xmax[i] += 0.1*xdiam[i];
      xdiam[i] *= 1.2;
    }
  }
  else
  {
    for (i = 0; i < DIM_OF_WORLD; i++)
    {
      xmin[i] = -1.0;
      xmax[i] =  1.0;
      xdiam[i] =  2.0;
    }
  }

  if (!title) title = "ALBERTAgraphics";

  if (!geom)
  {
    int size = 400, xsize, ysize;

    if (xdiam[0] >= xdiam[1])
    {
      xsize = size;
      ysize = size*xdiam[1]/xdiam[0];
    }
    else
    {
      xsize = size*xdiam[0]/xdiam[1];
      ysize = size;
    }
    sprintf(geometry, "%dx%d", xsize, ysize);
    geom = geometry;
  }

  if ((winO = OGL_create_window(title, geom)))
  {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glOrtho(xmin[0], xmax[0], xmin[1], xmax[1], -1.0, 1.0);

    return((GRAPH_WINDOW) winO);
  }

  return(nil);
}

void graph_clear_window(GRAPH_WINDOW win, const GRAPH_RGBCOLOR c)
{
  OGL_clear_window((OGL_WINDOW *)win, c);
  return;
}

void graph_mesh(GRAPH_WINDOW win, MESH *mesh, const GRAPH_RGBCOLOR c, FLAGS flag)
{
  FUNCNAME("graph_mesh");
  TRAVERSE_STACK  *stack;
  const EL_INFO   *el_info;
  REAL_D          world[N_VERTICES];
  REAL            coord[N_VERTICES][2];
  const float     *color;
  PARAMETRIC *parametric = nil;
  
  if (!win || !mesh)  return;

  if (DIM!=DIM_OF_WORLD) {
    ERROR_EXIT("How did you manage to open a window for DIM!=DIM_OF_WORLD?\n");
    return;
  }

  /* Standardfarbe setzen */
  color = c ? c : rgb_black;

  parametric = mesh->parametric;

  OGL_set_std_window((OGL_WINDOW *) win);

  glLineWidth(1);
  glPointSize(3);

  stack  = get_traverse_stack();
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS|FILL_BOUND);
  while (el_info)
  {
    glColor3fv(color);
    if (flag & GRAPH_MESH_ELEMENT_MARK)
    {
      if (el_info->el->mark > 0)
	glColor3fv(rgb_red);
      else if (el_info->el->mark < 0)
	glColor3fv(rgb_blue);
    }

    glBegin(GL_LINE_STRIP);
    if (parametric && parametric->init_element(el_info, parametric)) 
    {
      const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};

      parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);
    }
    else 
    {
      world[0][0] = el_info->coord[0][0];
      world[1][0] = el_info->coord[1][0];
    }
    coord[0][0] = world[0][0];
    coord[1][0] = world[1][0];
    coord[0][1] = coord[1][1] = 0.0;

    glVertex2dv(coord[0]);
    glVertex2dv(coord[1]);
    glEnd();
/*--------------------------------------------------------------------------*/
/*--- draw element boundaries                                            ---*/
/*--------------------------------------------------------------------------*/
    if (el_info->bound[0])
    {
      if (el_info->bound[0] > 0)
	glColor3fv(rgb_blue);
      else if (el_info->bound[0] < 0)
	glColor3fv(rgb_red);

      glPointSize(5);
      glBegin(GL_POINTS);
      glVertex2dv(coord[0]);
      glEnd();
    }

    if (el_info->bound[1] > 0)
      glColor3fv(rgb_blue);
    else if (el_info->bound[1] < 0)
      glColor3fv(rgb_red);
    else
      glColor3fv(color);

    glBegin(GL_POINTS);
    glVertex2dv(coord[1]);
    glEnd();

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);

  glFlush();
  return;
}

void graph1d_drv(GRAPH_WINDOW win, const DOF_REAL_VEC *uh,
		 REAL min, REAL max, int refine, const GRAPH_RGBCOLOR c)
{
  FUNCNAME("graph1d_drv");
  TRAVERSE_STACK  *stack;
  const EL_INFO   *el_info;
  MESH            *mesh;
  REAL_D          world[N_VERTICES];
  double          coord[2], scal;
  const REAL      *(*get_uh_el)(const EL *el, const DOF_REAL_VEC *, REAL *);
  PARAMETRIC      *parametric = nil;
  int             i,n_refine;
  static int      n_refines[6] = {1, 2, 4, 8, 16, 32};
  const BAS_FCTS  *bas_fcts;
  REAL            uh_val;

  if (!win || !uh)  return;

  if (DIM!=DIM_OF_WORLD) {
    ERROR_EXIT("How did you manage to open a window for DIM!=DIM_OF_WORLD?\n");
    return;
  }

  parametric = uh->fe_space->mesh->parametric;
  if ((parametric)&&(DIM!=DIM_OF_WORLD)) 
  {
    WARNING("This function is not tested for parametric elements\n"); 
    WARNING("with DIM!=DIM_OF_WORLD!\n");
  }

  if (min >= max)
  {
    min = dof_min(uh);
    max = dof_max(uh);
  }
  min = ABS(min);
  max = ABS(max);

  scal = min == max ? 0.9 : 0.9/MAX(min,max);

  bas_fcts = uh->fe_space->bas_fcts;
  if (refine < 0) refine = bas_fcts->degree-1;
  if (refine <= 0) 
      n_refine = 0;
  else if (refine > 5)
      refine = 5;
  n_refine = n_refines[refine];

  OGL_set_std_window((OGL_WINDOW *) win);

  glLineWidth(1);
  glColor3fv(c ? c : rgb_black);

  glBegin(GL_LINE_STRIP);
  mesh = uh->fe_space->mesh;
  get_uh_el = bas_fcts->get_real_vec;

  stack  = get_traverse_stack();
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while (el_info)
  {
    const REAL *uh_el = (*get_uh_el)(el_info->el, uh, nil);

    if (parametric && parametric->init_element(el_info, parametric)) 
    {
      const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};

      parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);
    }
    else 
    {
      world[0][0] = el_info->coord[0][0];
      world[1][0] = el_info->coord[1][0];
    }     

    coord[0] = world[0][0];
    coord[1] = scal*uh_el[0];
    glVertex2dv(coord);

    for (i=1; i<n_refine; ++i) {
	REAL lambda[DIM+1];
	lambda[1] = ((REAL)i)/((REAL)n_refine);
	lambda[0] = 1.0 - lambda[1];
	uh_val = eval_uh(lambda, uh_el, bas_fcts);
	coord[0] = lambda[0] * world[0][0] + lambda[1] * world[1][0];
	coord[1] = scal*uh_val;
	glVertex2dv(coord);
    }

    coord[0] = world[1][0];
    coord[1] = scal*uh_el[1];
    glVertex2dv(coord);

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);
  glEnd();
  glFlush();
  return;
}

void graph1d_drv_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *uh,
		   REAL min, REAL max, int refine, const GRAPH_RGBCOLOR c)
{
  FUNCNAME("graph1d_drv_d");
  TRAVERSE_STACK *stack;
  const EL_INFO  *el_info;
  MESH           *mesh;
  REAL_D         world[N_VERTICES];
  double         coord[2], scal;
  const REAL_D   *(*get_uh_d_el)(const EL*, const DOF_REAL_D_VEC *, REAL_D*);
  PARAMETRIC     *parametric = nil;
  int            i,n_refine;
  static int     n_refines[6] = {1, 2, 4, 8, 16, 32};
  const BAS_FCTS *bas_fcts;
  const REAL     *uh_val;

  if (!win || !uh)  return;

  if (DIM!=DIM_OF_WORLD) {
    ERROR_EXIT("How did you manage to open a window for DIM!=DIM_OF_WORLD?\n");
    return;
  }

  parametric = uh->fe_space->mesh->parametric;
  if ((parametric)&&(DIM!=DIM_OF_WORLD)) 
  {
    WARNING("This function is not tested for parametric elements\n"); 
    WARNING("with DIM!=DIM_OF_WORLD!\n");
  }

  if (min >= max)
  {
    min = dof_min_d(uh);
    max = dof_max_d(uh);
  }

  scal = min == max ? 0.9 : 0.9/max;

  bas_fcts = uh->fe_space->bas_fcts;
  if (refine < 0) refine = bas_fcts->degree-1;
  if (refine <= 0) 
      n_refine = 0;
  else if (refine > 5)
      refine = 5;
  n_refine = n_refines[refine];

  OGL_set_std_window((OGL_WINDOW *) win);

  glLineWidth(1);
  glColor3fv(c ? c : rgb_black);

  glBegin(GL_LINE_STRIP);
  mesh = uh->fe_space->mesh;
  get_uh_d_el = bas_fcts->get_real_d_vec;

  stack  = get_traverse_stack();
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while (el_info)
  {
    const REAL_D *uh_d_el = (*get_uh_d_el)(el_info->el, uh, nil);

    if (parametric && parametric->init_element(el_info, parametric)) 
    {
      const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};

      parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);
    }
    else 
    {
      world[0][0] = el_info->coord[0][0];
      world[1][0] = el_info->coord[1][0];
    }     

    coord[0] = world[0][0];
    coord[1] = scal*NORM_DOW(uh_d_el[0]);
    glVertex2dv(coord);

    for (i=1; i<n_refine; ++i) {
	REAL lambda[DIM+1];
	lambda[1] = ((REAL)i)/((REAL)n_refine);
	lambda[0] = 1.0 - lambda[1];
	uh_val = eval_uh_d(lambda, uh_d_el, bas_fcts, nil);
	coord[0] = lambda[0] * world[0][0] + lambda[1] * world[1][0];
	coord[1] = scal*NORM_DOW(uh_val);
	glVertex2dv(coord);
    }

    coord[0] = world[1][0];
    coord[1] = scal*NORM_DOW(uh_d_el[1]);
    glVertex2dv(coord);

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);
  glEnd();
  glFlush();
  return;
}

/*--------------------------------------------------------------------------*/

void graph1d_el_est(GRAPH_WINDOW win, MESH *mesh, REAL (*get_est)(EL *),
		    REAL min, REAL max, const GRAPH_RGBCOLOR c)
{
  FUNCNAME("graph1d_el_est");
  TRAVERSE_STACK  *stack;
  const EL_INFO   *el_info;
  double          coord[2], scal, est;
  REAL_D          world[N_VERTICES];
  PARAMETRIC      *parametric = nil;

  if (!win || !mesh || !get_est)  return;

  if (DIM!=DIM_OF_WORLD) {
    ERROR_EXIT("How did you manage to open a window for DIM!=DIM_OF_WORLD?\n");
    return;
  }

  parametric = mesh->parametric;
  if ((parametric)&&(DIM!=DIM_OF_WORLD)) 
  {
    WARNING("This function is not tested for parametric elements\n"); 
    WARNING("with DIM!=DIM_OF_WORLD!\n");
  }

  stack  = get_traverse_stack();

  if (min >= max)
  {
    el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL);
    while (el_info)
    {
      est = (*get_est)(el_info->el);
      min = MIN(min,est);
      max = MAX(max,est);

      el_info = traverse_next(stack, el_info);
    }
  }
  min = ABS(min);
  max = ABS(max);
  scal = min == max ? 0.9 : 0.9/MAX(min,max);

  OGL_set_std_window((OGL_WINDOW *) win);

  glLineWidth(1);
  glColor3fv(c ? c : rgb_black);

  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while (el_info)
  {
    est = (*get_est)(el_info->el);

    if (parametric && parametric->init_element(el_info, parametric)) 
    {
      const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};
      
      parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);
    }
    else 
    {
      world[0][0] = el_info->coord[0][0];
      world[1][0] = el_info->coord[1][0];
    } 
    
    glBegin(GL_LINE_STRIP);
    coord[0] = world[0][0];
    coord[1] = scal*est;
    glVertex2dv(coord);
    coord[0] = world[1][0];
    coord[1] = scal*est;
    glVertex2dv(coord);
    glEnd();

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);
  glFlush();
  return;
}

/*--------------------------------------------------------------------------*/

void graph1d_fvalues(GRAPH_WINDOW win, MESH *mesh,
		     REAL (*f)(const EL_INFO *, const REAL *lambda),
		     FLAGS fill_flag, REAL min, REAL max, int refine,
		     const GRAPH_RGBCOLOR c)
{
  FUNCNAME("graph1d_fvalues");
  TRAVERSE_STACK  *stack;
  const EL_INFO   *el_info;
  double          coord[2], scal, fval, delta, lam[2];
  REAL_D          world[N_VERTICES];
  PARAMETRIC      *parametric = nil;
  int             i, npts;

  if (!win || !mesh || !f)  return;

  if (DIM!=DIM_OF_WORLD) {
    ERROR_EXIT("How did you manage to open a window for DIM!=DIM_OF_WORLD?\n");
    return;
  }

  parametric = mesh->parametric;
  if ((parametric)&&(DIM!=DIM_OF_WORLD)) 
  {
    WARNING("This function is not tested for parametric elements\n"); 
    WARNING("with DIM!=DIM_OF_WORLD!\n");
  }

  stack  = get_traverse_stack();

  if (min >= max)
  {
#define NTEST 4
    static REAL b[NTEST][DIM+1] = {{1.0, 0.0}, {0.0, 1.0},
				   {0.667, 0.333}, {0.333,0.667}};
    int i, ntest;

    if (refine>0) ntest = NTEST;
    else          ntest = 2;
#undef NTEST
    
    min = max = 0.0;
    el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL | fill_flag);
    while (el_info)
    {
	for (i = 0; i < ntest; i++) {
	    fval = f(el_info, b[i]);
	    min = MIN(min,fval);
	    max = MAX(max,fval);
	}
      el_info = traverse_next(stack, el_info);
    }
  }
  min = ABS(min);
  max = ABS(max);
  scal = min == max ? 0.9 : 0.9/MAX(min,max);

  OGL_set_std_window((OGL_WINDOW *) win);

  npts = 1;
  delta = 1.0;
  if (refine > 0) {
      refine = MIN(refine, 8);
      for (i=0; i<refine; ++i) {
	  npts  *= 2;
	  delta /= 2;
      }
  }
  ++npts;

  glLineWidth(1);
  glColor3fv(c ? c : rgb_black);

  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS|fill_flag);
  while (el_info)
  {
    if (parametric && parametric->init_element(el_info, parametric)) 
    {
      const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};
      
      parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);
    }
    else 
    {
      parametric = nil;
      world[0][0] = el_info->coord[0][0];
      world[1][0] = el_info->coord[1][0];
    } 
    
    glBegin(GL_LINE_STRIP);
    for (i=0; i<npts; ++i) {
	lam[1] = i * delta;
	lam[0] = 1.0 - lam[1];
	fval = f(el_info, lam);
	coord[0] = lam[0] * world[0][0] + lam[1] * world[1][0];
	coord[1] = scal*fval;
	glVertex2dv(coord);
   }
    glEnd();

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);
  glFlush();
  return;
}

/*--------------------------------------------------------------------------*/

void graph_point(GRAPH_WINDOW win, const REAL p0[2], const GRAPH_RGBCOLOR c,
		 float ps)
{
  OGL_WINDOW *ogl_win = win;

  OGL_set_std_window(ogl_win);

  glColor3fv(c ? c : rgb_black);
  glPointSize(ps > 0 ? ps : 1.0);
    
  glBegin(GL_POINTS);
  glVertex2d(p0[0], p0[1]);
  glEnd();
  glFlush();
  return;
}

void graph_line(GRAPH_WINDOW win, const REAL p0[2], const REAL p1[2], 
		const GRAPH_RGBCOLOR c, float lw)
{
  float xy[2];
  int   j;

  OGL_set_std_window((OGL_WINDOW *) win);

  glColor3fv(c ? c : rgb_black);
  glLineWidth(lw > 0 ? lw : 1.0);

  glBegin(GL_LINE_STRIP);
  for (j = 0; j < 2; j++) xy[j] = p0[j];
  glVertex2fv(xy);
  for (j = 0; j < 2; j++) xy[j] = p1[j];
  glVertex2fv(xy);
  glEnd();
  glFlush();

  return;
}

/****************************************************************************/
/*   common graphic routines for all dimensions                             */
/****************************************************************************/

void graph_drv(GRAPH_WINDOW win, const DOF_REAL_VEC *u,
	       REAL min, REAL max, int refine)
{
    graph1d_drv(win, u, min, max, refine, nil);
}

void graph_drv_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *ud,
		 REAL min, REAL max, int refine)
{
    graph1d_drv_d(win, ud, min, max, refine, nil);
}

void graph_el_est(GRAPH_WINDOW win, MESH *mesh, REAL (*get_est)(EL *),
		  REAL min, REAL max)
{
     graph1d_el_est(win, mesh, get_est, min, max, nil);
}

void graph_fvalues(GRAPH_WINDOW win, MESH *mesh,
		   REAL (*f)(const EL_INFO *, const REAL *lambda),
		   FLAGS fill_flag, REAL min, REAL max, int refine)
{
     graph1d_fvalues(win, mesh, f, fill_flag, min, max, refine, nil);
}
