/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     refine.c                                                       */
/*                                                                          */
/*    !!!  includes ../Common/refine_common.c  !!!                          */
/*                                                                          */
/* description:  recursive refinement of 1 dim. hierarchical meshes;        */
/*               implementation of the newest vertex bisection              */
/*               file contains all routines depending on DIM == 1;          */
/*               routines for both dimensions in ../Common/refine_common.c  */
/*               are included at the end of the file                        */
/*                                                                          */
/* global variables that have to be defined for ../Common/refine_common.c   */
/*  static int            call_refine_interpol;                             */
/*  static int            do_more_refine;                                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

void free_dof(DOF *dof, MESH *mesh, int position);
void free_leaf_data(void *leaf_data, MESH *mesh);

static int   call_refine_interpol;
static int   do_more_refine; 
static void  refine_interpol(RC_LIST_EL *list, int n_el);

/*--------------------------------------------------------------------------*/
/*  refine_fct: bisects a single element into child[0] and child[1]         */
/*--------------------------------------------------------------------------*/

static void refine_fct(const EL_INFO *el_info)
{
  MESH       *mesh = el_info->mesh;
  EL         *el = el_info->el, *child[2];

  if (el->mark <= 0)
    return;

  child[0] = get_element(mesh);
  child[1] = get_element(mesh);
  child[0]->mark = child[1]->mark = MAX(0, el->mark-1);
  el->mark = 0;

/*--------------------------------------------------------------------------*/
/*  transfer hided data from parent to children                             */
/*--------------------------------------------------------------------------*/

  if (el->child[1]  &&  mesh->leaf_data_info->refine_leaf_data)
    mesh->leaf_data_info->refine_leaf_data(el, child);
  free_leaf_data((void *) el->child[1], mesh);

  el->child[0] = child[0];
  el->child[1] = child[1];
  
  if (child[0]->mark > 0) do_more_refine = true;

  child[0]->dof[1] = child[1]->dof[0] = get_dof(mesh, VERTEX);

/*--------------------------------------------------------------------------*/
/*  the other vertices are handed on from the parent                        */
/*--------------------------------------------------------------------------*/
  child[0]->dof[0] = el->dof[0];
  child[1]->dof[1] = el->dof[1];

/*--------------------------------------------------------------------------*/
/*  there is one more leaf element, two hierachical elements,               */
/*  one more vertex                                                         */
/*--------------------------------------------------------------------------*/

  mesh->n_elements++;
  mesh->n_vertices++;
  mesh->n_hier_elements += 2;

  if (mesh->n_dof[CENTER])
  {
/*--------------------------------------------------------------------------*/
/* there are dofs at the barycenter of the triangles                        */
/*--------------------------------------------------------------------------*/
    child[0]->dof[mesh->node[CENTER]] = get_dof(mesh, CENTER);
    child[1]->dof[mesh->node[CENTER]] = get_dof(mesh, CENTER);
  }

/*--------------------------------------------------------------------------*/
/*  if there are functions to interpolate data to the finer grid, do so     */
/*--------------------------------------------------------------------------*/
  
  if (call_refine_interpol)
  {
    RC_LIST_EL  ref_list[1] = {{nil, 0, 0}};
    ref_list->el = el;
    refine_interpol(ref_list, 1);
  }

  if (!mesh->preserve_coarse_dofs && mesh->n_dof[CENTER])
  {
    free_dof(el->dof[mesh->node[CENTER]], mesh, CENTER);
    el->dof[mesh->node[CENTER]] = nil;
  }

#if NEIGH_IN_EL
  {
    EL  *neigh;
    int ov;

    for (i_child = 0; i_child < 2; i_child++)
    {
/*--------------------------------------------------------------------------*/
/*  first neighbours handed on by the parent                                */
/*--------------------------------------------------------------------------*/
      if ((neigh = el->neigh[1-i_child]))
      {
	child[i_child]->neigh[1-i_child] = neigh;
	child[i_child]->opp_vertex[1-i_child] = i_child;
	neigh->neigh[i_child] = child[i_child];
	neigh->opp_vertex[i_child] = 1-i_child;
      }
/*--------------------------------------------------------------------------*/
/*  now child[0] to child[1] relationship                                   */
/*--------------------------------------------------------------------------*/
      child[i_child]->neigh[i_child] = child[1-i_child];
      child[i_child]->opp_vertex[1-i_child] = 1-i_child;
    }
  }
#endif

  return;
}

#include "refine_common.c"
