/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     refine.c                                                       */
/*                                                                          */
/*    !!!  includes ../Common/refine_common.c  !!!                          */
/*                                                                          */
/* description:  recursive refinement of 2 dim. hierarchical meshes;        */
/*               implementation of the newest vertex bisection              */
/*               file contains all routines depending on DIM == 2;          */
/*               routines for both dimensions in ../Common/refine_common.c  */
/*               are included at the end of the file                        */
/*                                                                          */
/* global variables that have to be defined for ../Common/refine_common.c   */
/*  static int            call_refine_interpol;                             */
/*  static int            do_more_refine;                                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

void free_dof(DOF *dof, MESH *mesh, int position);
void free_leaf_data(void *leaf_data, MESH *mesh);

static MESH          *refine_mesh;
static int           do_more_recursive_refine;
static int           new_coords;
static int           call_refine_interpol;

#if !NEIGH_IN_EL
static TRAVERSE_STACK *stack;
#endif

int vertex_of_edge[3][2] = {{1,2}, {2,0}, {0,1}};

static const EL_INFO  *refine_function(const EL_INFO *);

static int get_refine_patch(EL_INFO **el_info, DOF *edge[2], int dir, 
			    RC_LIST_EL *ref_list, int *n_neigh)
{
  FUNCNAME("get_refine_patch");
  EL       *el = (*el_info)->el;

#if NEIGH_IN_EL
  EL_INFO  neigh_info[1];
  EL       *neigh;
/*--------------------------------------------------------------------------*/
/*  get neighbour and check whether it is compatible devisible or not       */
/*--------------------------------------------------------------------------*/
  neigh = el->neigh[2];
  if (neigh && el->opp_vertex[2] != 2)
  {
    neigh->mark = MAX(neigh->mark, 1);
    neigh_info->el = neigh;
    refine_function(neigh_info);
  }
#else
  const EL_INFO  *neigh_info;
  int            opp_vertex;

  if (NEIGH(el,(*el_info))[2] && OPP_VERTEX(el,(*el_info))[2] != 2)
  {
/*--------------------------------------------------------------------------*/
/*  neighbour is not compatible devisible; refine neighbour first, store the*/
/*  opp_vertex to traverse back to el                                       */
/*--------------------------------------------------------------------------*/
    opp_vertex = OPP_VERTEX(el,(*el_info))[2];

    neigh_info = traverse_neighbour(stack, *el_info, 2);
    neigh_info->el->mark = MAX(neigh_info->el->mark, 1);
    neigh_info = refine_function(neigh_info);

/*--------------------------------------------------------------------------*/
/*  now go back to the original element and refine the patch                */
/*--------------------------------------------------------------------------*/
    neigh_info = traverse_neighbour(stack, neigh_info, opp_vertex);
    *el_info = (EL_INFO *) neigh_info;
    TEST_EXIT(neigh_info->el == el)("invalid traverse_neighbour1");
  }
#endif
  
  if ((ref_list[1].el = NEIGH(el,(*el_info))[2]))
  {
    TEST_EXIT(OPP_VERTEX(el,(*el_info))[2] == 2)
      ("no compatible ref. edge after recursive refinement of neighbour");
    *n_neigh = 2;
  }
  return(0);
}

/*--------------------------------------------------------------------------*/
/*  bisect_element: bisects a single element into child[0] and child[1]     */
/*--------------------------------------------------------------------------*/

static void  bisect_element(EL *el, DOF *dof[3])
{
  EL        *child[2];
  int       i_child;

  child[0] = get_element(refine_mesh);
  child[1] = get_element(refine_mesh);
  child[0]->mark = child[1]->mark = MAX(0, el->mark-1);
  el->mark = 0;

/*--------------------------------------------------------------------------*/
/*  transfer hided data from parent to children                             */
/*--------------------------------------------------------------------------*/

  if (el->child[1]  &&  refine_mesh->leaf_data_info->refine_leaf_data)
    refine_mesh->leaf_data_info->refine_leaf_data(el, child);
  free_leaf_data((void *) el->child[1], refine_mesh);

  el->child[0] = child[0];
  el->child[1] = child[1];
  
  if (child[0]->mark > 0) do_more_recursive_refine = true;

/*--------------------------------------------------------------------------*/
/*  vertex 2 is the newest vertex                                           */
/*--------------------------------------------------------------------------*/

  child[0]->dof[2] = child[1]->dof[2] = dof[0];

/*--------------------------------------------------------------------------*/
/*  the other vertices are handed on from the parent                        */
/*--------------------------------------------------------------------------*/

  for (i_child = 0; i_child < 2; i_child++)
  {
    child[i_child]->dof[i_child] = el->dof[2];
    child[i_child]->dof[1-i_child] = el->dof[i_child];
  }

/*--------------------------------------------------------------------------*/
/*  there is one more leaf element, two hierachical elements and one more   */
/*  edge                                        			    */
/*--------------------------------------------------------------------------*/

  refine_mesh->n_edges++;
  refine_mesh->n_elements++;
  refine_mesh->n_hier_elements += 2;

  if (refine_mesh->n_dof[EDGE])
  {
/*--------------------------------------------------------------------------*/
/*  there are dof's in the midpoint of the edges			    */
/*--------------------------------------------------------------------------*/
    child[0]->dof[4] = child[1]->dof[3] = get_dof(refine_mesh, EDGE);

/*--------------------------------------------------------------------------*/
/*  dofs handed on by the parent                                            */
/*--------------------------------------------------------------------------*/
    child[0]->dof[5] = el->dof[4];
    child[1]->dof[5] = el->dof[3];

/*--------------------------------------------------------------------------*/
/*  dofs in the refinement edge                                             */
/*--------------------------------------------------------------------------*/
    child[0]->dof[3] = dof[1];
    child[1]->dof[4] = dof[2];
  }

  if (refine_mesh->n_dof[CENTER])
  {
    int   node = refine_mesh->node[CENTER];

/*--------------------------------------------------------------------------*/
/* there are dofs at the barycenter of the triangles                        */
/*--------------------------------------------------------------------------*/
    child[0]->dof[node] = get_dof(refine_mesh, CENTER);
    child[1]->dof[node] = get_dof(refine_mesh, CENTER);
  }

#if NEIGH_IN_EL
  {
    EL  *neigh;
    int ov;

    for (i_child = 0; i_child < 2; i_child++)
    {
/*--------------------------------------------------------------------------*/
/*  first neighbours handed on by the parent                                */
/*--------------------------------------------------------------------------*/
      if (neigh = child[i_child]->neigh[2] = el->neigh[1 - i_child])
      {
	ov = el->opp_vertex[1 - i_child];
	neigh->opp_vertex[ov] = 2;
	neigh->neigh[ov] = child[i_child];
	child[i_child]->opp_vertex[2] = ov;
      }
/*--------------------------------------------------------------------------*/
/*  now child[0] to child[1] relationship                                   */
/*--------------------------------------------------------------------------*/
      child[i_child]->neigh[1-i_child] = child[1-i_child];
      child[i_child]->opp_vertex[1-i_child] = i_child;
    }
  }
#endif

  return;
}

/*--------------------------------------------------------------------------*/
/*  if DOFs for higher order should be removed, this is done by             */
/*  remove_dof_parent()                                                     */
/*--------------------------------------------------------------------------*/

static void remove_dof_parents(RC_LIST_EL ref_list[], int n_neigh)
{
  int    i, j, node;

  if (refine_mesh->n_dof[EDGE])
  {  
    node = refine_mesh->node[EDGE];

    for (i = 0; i < n_neigh; i++)
    {
      for (j = 0; j < N_EDGES; j++)
	ref_list[i].el->dof[node+j] = nil;
    }
  }

  if (refine_mesh->n_dof[CENTER])
  {
    node = refine_mesh->node[CENTER];
    for (i = 0; i < n_neigh; i++)
    {
      free_dof(ref_list[i].el->dof[node], refine_mesh, CENTER);
      ref_list[i].el->dof[node] = nil;
    }
  }
  return;
}
/*--------------------------------------------------------------------------*/

static void     refine_interpol(RC_LIST_EL *list, int n_el);

static void  bisect_patch(DOF *edge[2], RC_LIST_EL ref_list[], int n_neigh, 
			  int bound)
{
  DOF      *dof[3] = {nil, nil, nil};
  EL       *el = ref_list[0].el, *neigh = ref_list[1].el;

/*--------------------------------------------------------------------------*/
/*  there is one new vertex in the refinement edge                          */
/*--------------------------------------------------------------------------*/

  dof[0] = get_dof(refine_mesh, VERTEX);

  refine_mesh->n_vertices++;
  refine_mesh->n_edges++;

  if (refine_mesh->n_dof[EDGE])
  {
/*--------------------------------------------------------------------------*/
/*  there are two additional dofs in the refinement edge                    */
/*--------------------------------------------------------------------------*/
    dof[1] = get_dof(refine_mesh, EDGE);
    dof[2] = get_dof(refine_mesh, EDGE);
  }

/*--------------------------------------------------------------------------*/
/*  first refine the element                                                */
/*--------------------------------------------------------------------------*/
  bisect_element(el, dof);

  if (neigh)
  {
    DOF    *tmp = dof[1];
/*--------------------------------------------------------------------------*/
/*  there is a neighbour; refine it also, but first exchange dof[1] and     */
/*  dof[2]; thus, dof[1] is always added on child[0]!                       */
/*--------------------------------------------------------------------------*/
    dof[1] = dof[2];
    dof[2] = tmp;

    bisect_element(neigh, dof);
  }
  else
  {
    new_coords++;
  }

#if NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  set neighbour information across refinement edge                        */
/*--------------------------------------------------------------------------*/
  if (neigh)
  {
    int   i_child;

    for (i_child = 0; i_child < 2; i_child++)
    {
      el->child[i_child]->neigh[i_child] = neigh->child[1-i_child];
      el->child[i_child]->opp_vertex[i_child] = 1-i_child;

      neigh->child[i_child]->neigh[i_child] = el->child[1-i_child];
      neigh->child[i_child]->opp_vertex[i_child] = 1-i_child;
    }
  }
  else
  {
    el->child[0]->neigh[0] = el->child[1]->neigh[1] = nil;
  }
#endif
  
/*--------------------------------------------------------------------------*/
/*  if there are functions to interpolate data to the finer grid, do so     */
/*--------------------------------------------------------------------------*/
  
  if (call_refine_interpol)  refine_interpol(ref_list, n_neigh);

  if (!refine_mesh->preserve_coarse_dofs)
  {
/*--------------------------------------------------------------------------*/
/*  if there should be no dof information on interior leaf elements remove  */
/*  dofs from edges and the centers of parents                              */
/*--------------------------------------------------------------------------*/
    if (refine_mesh->n_dof[EDGE])
    {
      int   node = refine_mesh->node[EDGE];

/*--------------------------------------------------------------------------*/
/*  the only DOF that can be freed is that in the refinement edge; all other*/
/*  DOFs are handed on the children                                         */
/*--------------------------------------------------------------------------*/

      free_dof(el->dof[node+2], refine_mesh, EDGE);
    }
    if (refine_mesh->n_dof[EDGE]  ||  refine_mesh->n_dof[CENTER])
      remove_dof_parents(ref_list, n_neigh);
  }

  return;
}

#include "refine_common.c"
