/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     traverse_r.c                                                   */
/*                                                                          */
/* description:                                                             */
/*           recursive mesh traversal - 2d routines:                        */
/*           fill_macro_info(), fill_elinfo()                               */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"
#include "traverse_r_common.c"

/*--------------------------------------------------------------------------*/
/*   fill_macro_info:                                		            */
/*   ----------------                                   		    */
/*   fill EL_INFO structure for a macro element         		    */
/*--------------------------------------------------------------------------*/

void fill_macro_info(MESH *mesh, const MACRO_EL *mel, EL_INFO *elinfo)
{
  FUNCNAME("fill_macro_info");
  EL       *nb;
  MACRO_EL *mnb;
  int      i, j, k;
#if !NEIGH_IN_EL
  FLAGS    fill_opp_coords;
#endif  

  elinfo->mesh     = mesh;
  elinfo->macro_el = mel;
  elinfo->el       = mel->el;
  elinfo->parent   = nil;
  elinfo->level    = 0;

  if (elinfo->fill_flag & FILL_COORDS) {

    for (i=0; i<N_VERTICES; i++) {
      if (mel->coord[i]) {
	for (j=0;j<DIM_OF_WORLD;j++) elinfo->coord[i][j] = mel->coord[i][j];
      }
      else {
	TEST_EXIT(0)("no mel->coord[%d]\n",i);
      }
    }
  }

#if NEIGH_IN_EL
  if (elinfo->fill_flag & FILL_OPP_COORDS) {

    for (i = 0; i < N_NEIGH; i++) {
      if ((mnb = mel->neigh[i])) {
	nb = mel->neigh[i]->el;
	k  = mel->opp_vertex[i];
	if (nb->child[0] && (k != 2)) {         /* make nb nearest el. */
	  if (nb->new_coord)
	  {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[i][j] = nb->new_coord[j];
	  }
	  else
	  {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[i][j] = 
		(mnb->coord[0][j] + mnb->coord[1][j]) / 2;
	  }
	}
	else {
	  for (j = 0; j < DIM_OF_WORLD; j++)
	  {
	    elinfo->opp_coord[i][j] = mnb->coord[k][j];
	  }
	}
      }
    }
  }

#else  /* ! NEIGH_IN_EL */

  if (elinfo->fill_flag & (FILL_OPP_COORDS | FILL_NEIGH)) {

    fill_opp_coords = (elinfo->fill_flag & FILL_OPP_COORDS);
    for (i = 0; i < N_NEIGH; i++) 
    {
      if ((mnb = mel->neigh[i])) 
      {
	nb = elinfo->neigh[i]      = mel->neigh[i]->el;
	k  = elinfo->opp_vertex[i] = mel->opp_vertex[i];

	if (nb->child[0] && (k != 2))  /*---  make nb nearest element  ---*/
	{
	  if  (fill_opp_coords) 
	  {
	    if (nb->new_coord)
	    {
	      for (j = 0; j < DIM_OF_WORLD; j++)
		elinfo->opp_coord[i][j] = nb->new_coord[j];
	    }
	    else
	    {
	      for (j = 0; j < DIM_OF_WORLD; j++) 
		elinfo->opp_coord[i][j] = 
		  (mnb->coord[0][j] + mnb->coord[1][j]) / 2;
	    }
	  }
	  nb = elinfo->neigh[i]      = nb->child[1-k];
	  k  = elinfo->opp_vertex[i] = 2;
	}
	else {
	  if  (fill_opp_coords) {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[i][j] = mnb->coord[k][j];
	  }
	}
      }
      else {
	elinfo->neigh[i] = nil;
      }
    }
  }
#endif

  if (elinfo->fill_flag & FILL_BOUND) {

    for (i = 0; i < N_VERTICES; i++)
      elinfo->bound[i] = mel->bound[i];

    for (i=0;i<N_NEIGH; i++)
      elinfo->boundary[i] = mel->boundary[i];

  }

}


/*--------------------------------------------------------------------------*/
/*   fill_elinfo:                                    		            */
/*   ------------                                       		    */
/*   fill EL_INFO structure for one child of an element   		    */
/*--------------------------------------------------------------------------*/

void fill_elinfo(int ichild, const EL_INFO *elinfo_old, EL_INFO *elinfo)
{
  FUNCNAME("fill_elinfo");
  int     j;
  EL      *nb;
  EL      *el = elinfo_old->el;
  FLAGS   fill_flag = elinfo_old->fill_flag;
#if !NEIGH_IN_EL
  FLAGS   fill_opp_coords;
#endif

  TEST_EXIT(el->child[0])("no children?\n");
  TEST_EXIT((elinfo->el = el->child[ichild]))("missing child %d?\n", ichild);

  elinfo->macro_el  = elinfo_old->macro_el;
  elinfo->fill_flag = fill_flag;
  elinfo->mesh      = elinfo_old->mesh;
  elinfo->parent    = el;
  elinfo->level     = elinfo_old->level + 1;


  if (fill_flag & FILL_COORDS)
  {
    if (el->new_coord) {
      for (j = 0; j < DIM_OF_WORLD; j++)
	elinfo->coord[2][j] = el->new_coord[j];
    }
    else {
      for (j = 0; j < DIM_OF_WORLD; j++)
	elinfo->coord[2][j] = 
	  0.5 * (elinfo_old->coord[0][j] + elinfo_old->coord[1][j]);
    }

    if (ichild==0) {
      for (j = 0; j < DIM_OF_WORLD; j++) {
	elinfo->coord[0][j] = elinfo_old->coord[2][j];
	elinfo->coord[1][j] = elinfo_old->coord[0][j];
      }
    }
    else {
      for (j = 0; j < DIM_OF_WORLD; j++) {
	elinfo->coord[0][j] = elinfo_old->coord[1][j];
	elinfo->coord[1][j] = elinfo_old->coord[2][j];
      }
    }
  }



#if NEIGH_IN_EL

  if (fill_flag & FILL_OPP_COORDS)
  {
    if (ichild == 0) 
    {
      if (el->neigh[1]) 
      {
	for (j = 0; j < DIM_OF_WORLD; j++)
	  elinfo->opp_coord[2][j] = elinfo_old->opp_coord[1][j];
      }

      if (el->child[0]  &&  el->child[1]->child[0]  &&
	  el->child[1]->child[1]) 
      {  
	if (el->child[1]->new_coord)
	{
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = el->child[1]->new_coord[j];
	}
	else
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = 
	      0.5 * (elinfo_old->coord[1][j] + elinfo_old->coord[2][j]);
	}
      }
      else 
      {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[1][j] = elinfo_old->coord[1][j];
      }

      if ((nb = el->neigh[2]))
      {
	TEST_EXIT(nb->child[0])("missing child?\n");
	nb = nb->child[1];
	if (nb->child[0]) 
	{
	  if (nb->new_coord) 
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = nb->new_coord[j];
	  }
	  else
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] =
		0.5 * (elinfo_old->opp_coord[2][j] + elinfo_old->coord[0][j]);
	  }
	}
	else 
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = elinfo_old->opp_coord[2][j];
	}
      }
    }
    else  /*---  now, for child[1]  ----------------------------------------*/
    {

      if (el->neigh[0]) 
      {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[2][j] = elinfo_old->opp_coord[0][j];
      }

      if (el->child[0]->child[0]) {
	if (el->child[0]->new_coord)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = el->child[0]->new_coord[j];
	else
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = 
	      0.5 * (elinfo_old->coord[0][j] + elinfo_old->coord[2][j]);
      }
      else {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[0][j] = elinfo_old->coord[0][j];
      }

      if ((nb = el->neigh[2]))
      {
	TEST((nb = nb->child[0]))("missing child?\n");
	if (nb->child[0]) 
	{
	  if (nb->new_coord)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = nb->new_coord[j];
	  else
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] =
		0.5 * (elinfo_old->opp_coord[2][j] + elinfo_old->coord[1][j]);
	}
	else {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = elinfo_old->opp_coord[2][j];
	}
      }
    }
  }

#else  /* ! NEIGH_IN_EL */

  if (fill_flag & (FILL_NEIGH | FILL_OPP_COORDS))
  {
    fill_opp_coords = (fill_flag & FILL_OPP_COORDS);
    if (ichild==0) 
    {
      if ((elinfo->neigh[2] = elinfo_old->neigh[1])) 
      {
	if (fill_opp_coords) 
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[2][j] = elinfo_old->opp_coord[1][j];
	}
      }
      elinfo->opp_vertex[2] = elinfo_old->opp_vertex[1];

      if (el->child[0]  &&  el->child[1]->child[0]  &&  
	  el->child[1]->child[0])
      {
	TEST_EXIT((elinfo->neigh[1] = el->child[1]->child[1]))
	  ("el->child[1]->child[0]!=nil, but el->child[1]->child[1]=nil\n");
	elinfo->opp_vertex[1] = 2;
	if (fill_opp_coords) 
	{
	  if (el->child[1]->new_coord)
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = el->child[1]->new_coord[j];
	  }
	  else
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = 
		0.5 * (elinfo_old->coord[1][j] + elinfo_old->coord[2][j]);
	  }
	}
      }
      else 
      {
	TEST_EXIT((elinfo->neigh[1] = el->child[1]))
	  ("el->child[0] != nil, but el->child[1] = nil\n");
	elinfo->opp_vertex[1] = 0;
	if (fill_opp_coords) 
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = elinfo_old->coord[1][j];
	}
      }

      if ((nb = elinfo_old->neigh[2]))
      {
	TEST(elinfo_old->opp_vertex[2] == 2)("invalid neighbour\n"); 
	TEST_EXIT(nb->child[0])("missing children?\n");
	TEST_EXIT((nb = nb->child[1]))("missing child[1]?\n");

	if (nb->child[0]) 
	{
	  elinfo->opp_vertex[0] = 2;
	  if (fill_opp_coords)
	  {
	    if (nb->new_coord)
	    {
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[0][j] = nb->new_coord[j];
	    }
	    else
	    {
	      for (j=0; j<DIM_OF_WORLD; j++)
	      {
		elinfo->opp_coord[0][j] =
		  0.5*(elinfo_old->opp_coord[2][j] + elinfo_old->coord[0][j]);
	      }
	    }
	  }    
	  nb = nb->child[0];
	}
	else 
	{
	  elinfo->opp_vertex[0] = 1;
	  if (fill_opp_coords) 
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = elinfo_old->opp_coord[2][j];
	  }
	}
      }
      elinfo->neigh[0] = nb;
    }
    else   /* ichild==1 */
    { 
      if ((elinfo->neigh[2] = elinfo_old->neigh[0]))
      {
	if (fill_opp_coords)
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[2][j] = elinfo_old->opp_coord[0][j];
	}
      }
      elinfo->opp_vertex[2] = elinfo_old->opp_vertex[0];

      if (el->child[0]->child[0])
      {
	elinfo->neigh[0] = el->child[0]->child[0];
	elinfo->opp_vertex[0] = 2;
	if (fill_opp_coords)
	{
	  if (el->child[0]->new_coord)
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = el->child[0]->new_coord[j];
	  }
	  else
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = 
		0.5 * (elinfo_old->coord[0][j] + elinfo_old->coord[2][j]);
	  }
	}
      }
      else 
      {
	elinfo->neigh[0] = el->child[0];
	elinfo->opp_vertex[0] = 1;
	if (fill_opp_coords)
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = elinfo_old->coord[0][j];
	}
      }

      if ((nb = elinfo_old->neigh[2]))
      {
	TEST(elinfo_old->opp_vertex[2] == 2)("invalid neighbour\n"); 
	TEST((nb = nb->child[0]))("missing child?\n");
	if (nb->child[0])
	{
	  elinfo->opp_vertex[1] = 2;
	  if (fill_opp_coords) 
	  {
	    if (nb->new_coord)
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[1][j] = nb->new_coord[j];
	    else
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[1][j] =
		  0.5*(elinfo_old->opp_coord[2][j] + elinfo_old->coord[1][j]);
	  }
	  nb = nb->child[1];
	}
	else 
	{
	  elinfo->opp_vertex[1] = 0;
	  if (fill_opp_coords) {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = elinfo_old->opp_coord[2][j];
	  }
	}
      }
      elinfo->neigh[1] = nb;
    }
  }
#endif


  if (fill_flag & FILL_BOUND)
  {
    if (elinfo_old->boundary[2])
      elinfo->bound[2] = elinfo_old->boundary[2]->bound;
    else
      elinfo->bound[2] = INTERIOR;

    if (ichild==0) {
      elinfo->bound[0] = elinfo_old->bound[2];
      elinfo->bound[1] = elinfo_old->bound[0];
      elinfo->boundary[0] = elinfo_old->boundary[2];
      elinfo->boundary[1] = nil;
      elinfo->boundary[2] = elinfo_old->boundary[1];
    }
    else {
      elinfo->bound[0] = elinfo_old->bound[1];
      elinfo->bound[1] = elinfo_old->bound[2];
      elinfo->boundary[0] = nil;
      elinfo->boundary[1] = elinfo_old->boundary[2];
      elinfo->boundary[2] = elinfo_old->boundary[0];
    }
  }

}



