/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     coarsen.c                                                      */
/*                                                                          */
/*    !!!  includes ../Common/coarsen_common.c  !!!                         */
/*                                                                          */
/* description:  coarsening of 3 dim. hierarchical meshes;                  */
/*               at the end of file Common/coarsen_common.c is included;    */
/*               file contains all routines depending on DIM == 3;          */
/*               routines for both dimensions in ../Common/coarsen_common.c */
/*               are included at the end of the file                        */
/*                                                                          */
/* global variables that have to be defined for Common/coarsen_common.c:    */
/*  static MESH           *coarse_mesh;                                     */
/*  static int            do_more;                                          */
/*  static int            call_coarse_restrict;                             */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/****************************************************************************/
/*  global variables used for coarsening                                    */
/****************************************************************************/

static MESH           *coarse_mesh;
static int            do_more;
static int            call_coarse_restrict;

#if !NEIGH_IN_EL
static TRAVERSE_STACK *stack;
#endif

/****************************************************************************/
/*  n_child_edge[el_type][ichild][dir]                                      */
/*  gives local index of new edge on child[ichild] part of face [2+dir] on  */
/*  the parent                                                              */
/****************************************************************************/
static U_CHAR  n_child_edge[3][2][2] = {{{5,4},{4,5}},{{5,4},{5,4}},
					{{5,4},{5,4}}};

/****************************************************************************/
/*  n_child_face[el_type][ichild][dir]                                      */
/*  gives local index of sub-face on child[ichild] part of face [2+dir] on  */
/*  the parent                                                              */
/****************************************************************************/
static U_CHAR  n_child_face[3][2][2] = {{{1,2},{2,1}},{{1,2},{1,2}},
					{{1,2},{1,2}}};

/****************************************************************************/
/*  coarse_element:  coarses a single element of the coarsening patch; dof's*/
/*  in the interior of the element are removed and dof's in the faces of the*/
/*  element are removed if neighbour has been coarsend or if the face is    */
/*  part of the domains boundary                                            */
/*                                                                          */
/*  if NEIGH_IN_EL neighbour information to elements outside the patch is   */
/*  set on the parent; neighbour information within the patch is still      */
/*  vailid                                                                  */
/****************************************************************************/

static void coarse_element(RC_LIST_EL *coarse_list)
{
  EL       *el = coarse_list->el, *child[2];
  EL       *neigh;
  int      dir, el_type, i, node;
#if NEIGH_IN_EL
  int      opp_v;
#endif

  child[0] = el->child[0]; child[1] = el->child[1];
  el_type = coarse_list->el_type;

#if NEIGH_IN_EL
/****************************************************************************/
/*  fill information about neighbours at the boundary of the coarsening     */
/*  patch; dofs for higher order elements in these faces still belong to    */
/*  the parent. Do not remove them form the mesh!!!                         */
/****************************************************************************/

  neigh = el->neigh[0] = child[1]->neigh[3];
  if (neigh  &&  neigh != el)
  {
    opp_v = el->opp_vertex[0] = child[1]->opp_vertex[3];
    neigh->neigh[opp_v] = el;
    neigh->opp_vertex[opp_v] = 0;
  }

  neigh = el->neigh[1] = child[0]->neigh[3];
  if (neigh  &&  neigh != el)
  {
    opp_v = el->opp_vertex[1] = child[0]->opp_vertex[3];
    neigh->neigh[opp_v] = el;
    neigh->opp_vertex[opp_v] = 1;
  }
#endif

/****************************************************************************/
/*  Information about patch neighbours is still valid! But edge and face    */
/*  dof's in a common face of patch neighbours have to be removed           */
/****************************************************************************/

  for (dir = 0; dir < 2; dir++)
  {
    neigh = coarse_list->neigh[dir] ? coarse_list->neigh[dir]->el : nil;
/*     opp_v = coarse_list->opp_vertex[dir]; */

    if (!neigh  ||  !neigh->child[0])
    {
/****************************************************************************/
/*  boundary face or  neigh has been coarsend: free the dof's in the face   */
/****************************************************************************/

      if (coarse_mesh->n_dof[EDGE])
      {
	node = coarse_mesh->node[EDGE] + n_child_edge[el_type][0][dir];
	free_dof(child[0]->dof[node], coarse_mesh, EDGE);
      }
      if (coarse_mesh->n_dof[FACE])
      {
	node = coarse_mesh->node[FACE] + n_child_face[el_type][0][dir];
	free_dof(child[0]->dof[node], coarse_mesh, FACE);
	node = coarse_mesh->node[FACE] + n_child_face[el_type][1][dir];
	free_dof(child[1]->dof[node], coarse_mesh, FACE);
      }
    }
  }  /*       for (dir = 0; dir < 2; dir++)                                 */

/****************************************************************************/
/*  finally remove the interior dof's: in the common face of child[0] and   */
/*  child[1] and at the two barycenter                                      */
/****************************************************************************/

  if (coarse_mesh->n_dof[FACE])
  {
    node = coarse_mesh->node[FACE];
    free_dof(child[0]->dof[node], coarse_mesh, FACE);
  }


  if (coarse_mesh->n_dof[CENTER])
  {
    node = coarse_mesh->node[CENTER];
    for (i = 0; i < 2; i++)
      free_dof(child[i]->dof[node], coarse_mesh, CENTER);
  }

/****************************************************************************/
/*  get new data on parent and transfer data from children to parent        */
/****************************************************************************/

  el->child[0] = nil;
  el->child[1] = (EL *) get_leaf_data(coarse_mesh);

  if (el->child[1]  &&  coarse_mesh->leaf_data_info->coarsen_leaf_data)
    coarse_mesh->leaf_data_info->coarsen_leaf_data(el, child);

  free_element(child[0], coarse_mesh);
  free_element(child[1], coarse_mesh);

  el->mark++;
  
  coarse_mesh->n_elements--;
  coarse_mesh->n_hier_elements -= 2;

  return;
}

/****************************************************************************/
/*  add_dof_parent: adds those dof's on the parent that are handed on by the*/
/*  children and adds the dof in the midpoint of the coarsening edge        */
/****************************************************************************/

static void add_dof_parent(RC_LIST_EL *coarse_list, DOF *dof)
{
  EL         *el = coarse_list->el;
  RC_LIST_EL *neigh;
  int        node;

  if (coarse_mesh->n_dof[EDGE])
  {
    node = coarse_mesh->node[EDGE];

/****************************************************************************/
/*  set the dof in the coarsening edge                                      */
/****************************************************************************/

    el->dof[node] = dof;

/****************************************************************************/
/*  and now those handed on by the children                                 */
/****************************************************************************/

    el->dof[node+1] = el->child[0]->dof[node];
    el->dof[node+2] = el->child[0]->dof[node+1];
    el->dof[node+5] = el->child[0]->dof[node+3];
    
    if (coarse_list->el_type)
    {
      el->dof[node+3] = el->child[1]->dof[node];
      el->dof[node+4] = el->child[1]->dof[node+1];
    }
    else
    {
      el->dof[node+3] = el->child[1]->dof[node+1];
      el->dof[node+4] = el->child[1]->dof[node];
    }
  }

  if (coarse_mesh->n_dof[FACE])
  {
    node = coarse_mesh->node[FACE];
/****************************************************************************/
/*  dof's at the faces within the patch: add new dof's if it is a boundary  */
/*  face or neighbour is an element behind this element in the corsen list  */
/****************************************************************************/
    
    neigh = coarse_list->neigh[0];
    if ((!neigh) || (neigh > coarse_list))
    {
      if (!el->dof[node+2])
      {
	el->dof[node+2] = get_dof(coarse_mesh, FACE);           /* face  2 */
	if (neigh)
	  neigh->el->dof[node+coarse_list->opp_vertex[0]] = el->dof[node+2];
      }
    }
    neigh = coarse_list->neigh[1];
    if ((!neigh) || (neigh > coarse_list))
    {
      if (!el->dof[node+3])
      {
	el->dof[node+3]  = get_dof(coarse_mesh, FACE);          /* face  3 */
	if (neigh)
	  neigh->el->dof[node+coarse_list->opp_vertex[1]] = el->dof[node+3];
      }
    }
/****************************************************************************/
/*  and now those handed on by the children                                 */
/****************************************************************************/

    el->dof[node  ] = el->child[1]->dof[node+3];
    el->dof[node+1] = el->child[0]->dof[node+3];
  }
  
  if (coarse_mesh->n_dof[CENTER])
  {
    node = coarse_mesh->node[CENTER];
    if (!el->dof[node])
      el->dof[node] = get_dof(coarse_mesh, CENTER);
  }
  return;
}


/****************************************************************************/
/*  get_coarse_patch:  gets the patch for coarsening starting on element    */
/*  el_info->el in direction of neighbour [3-dir]; returns 1 if a boundary  */
/*  reached and 0 if we come back to the starting element.                  */
/*                                                                          */
/*  if NEIGH_IN_EL we only can find the complete coarsening patch if the    */
/*  can be coarsend; otherwise neighbour information is not valid for       */
/*  parents; in such situation we stop looping around the edge and return 0 */
/*                                                                          */
/*  if !NEIGH_IN_EL we complete the loop also in the case of a incompatible */
/*  coarsening patch since then all marks of patch elements are reset by    */
/*  do_coarse_patch() and this minimizes calls of traverse_neighbour();     */
/*  if we reach a boundary while looping around the edge we loop back to    */
/*  the starting element before we return                                   */
/****************************************************************************/
 
static int get_coarse_patch(const EL_INFO *el_info, DOF *edge[2], int dir,
			    RC_LIST_EL coarse_list[], int *n_neigh)
{
  FUNCNAME("get_coarse_patch");
#if !NEIGH_IN_EL
  const EL_INFO *neigh_info;
#endif
  EL        *el, *neigh;
  int       i, j, k, opp_v, edge_no;

  static U_CHAR  next_el[6][2] = {{3,2},{1,3},{1,2},{0,3},{0,2},{0,1}};

  el = el_info->el;

  if ((neigh = NEIGH(el,el_info)[3-dir]) == nil)  return(1);

  opp_v = OPP_VERTEX(el,el_info)[3-dir];

#if !NEIGH_IN_EL
  neigh_info = traverse_neighbour(stack, el_info, 3-dir);
  TEST_EXIT(neigh == neigh_info->el)
    ("neigh %d and neigh_info->el %d are not identical\n", 
     INDEX(neigh), INDEX(neigh_info->el));
/****************************************************************************/
/*  we have to go back to the starting element via opp_v values             */
/*  correct information is produce by get_neigh_on_patch()                  */
/****************************************************************************/
  coarse_list[*n_neigh].opp_vertex[0] = opp_v;  
#endif
  coarse_list[*n_neigh].el = neigh;
  coarse_list[*n_neigh].el_type = EL_TYPE(neigh,neigh_info);

  while (neigh != el)
  {
    for (j = 0; j < N_VERTICES; j++)
      if (neigh->dof[j] == edge[0])  break;
    for (k = 0; k < N_VERTICES; k++)
      if (neigh->dof[k] == edge[1])  break;

    TEST_EXIT(j < N_VERTICES  &&  k < N_VERTICES)
      ("dof %d or dof %d not found on element %d with nodes (%d %d %d %d)\n", 
       edge[0][0], edge[1][0], INDEX(neigh), neigh->dof[0][0],
       neigh->dof[1][0], neigh->dof[2][0], neigh->dof[3][0]);

    edge_no = edge_of_dofs[j][k];
    coarse_list[*n_neigh].flag = edge_no == 0;

#if NEIGH_IN_EL
/****************************************************************************/
/*  neigh is not compatible for coarsening and neighbour information is not */
/*  valid for further looping around the edge                               */
/****************************************************************************/

    if (edge_no)
    {
      ++*n_neigh;
      return(0);
    }
#endif

/****************************************************************************/
/*  get the direction of the next neighbour				    */
/****************************************************************************/

    if (next_el[edge_no][0] != opp_v)
      i = next_el[edge_no][0];
    else
      i = next_el[edge_no][1];

    ++*n_neigh;

    opp_v = OPP_VERTEX(neigh,neigh_info)[i];
    if ((neigh = NEIGH(neigh,neigh_info)[i]))
    {
#if !NEIGH_IN_EL
      neigh_info = traverse_neighbour(stack, neigh_info, i);
      TEST_EXIT(neigh == neigh_info->el)
	("neigh %d and neigh_info->el %d are not identical\n", 
	 INDEX(neigh), INDEX(neigh_info->el));
/****************************************************************************/
/*  we have to go back to the starting element via opp_v values             */
/*  correct information is produce by get_neigh_on_patch()                  */
/****************************************************************************/
      coarse_list[*n_neigh].opp_vertex[0] = opp_v;  
#endif
      coarse_list[*n_neigh].el = neigh;
      coarse_list[*n_neigh].el_type = EL_TYPE(neigh,neigh_info);
    }
    else
      break;
  }

  if (neigh == el)  
    return(0);

#if !NEIGH_IN_EL
/****************************************************************************/
/*  the domain's boundary is reached; loop back to the starting el          */
/****************************************************************************/

  i = *n_neigh-1;
  opp_v = coarse_list[i].opp_vertex[0];
  do
  {
    TEST_EXIT(NEIGH(neigh,neigh_info)[opp_v]  &&  i > 0)
      ("while looping back domains boundary was reached or i == 0\n");
    opp_v = coarse_list[i--].opp_vertex[0];
    neigh_info = traverse_neighbour(stack, neigh_info, opp_v);
  } while(neigh_info->el != el);
#endif

  return(1);
}


/****************************************************************************/
/*  coarse_patch: first rebuild the dofs on the parents then do restriction */
/*  of data (if possible) and finally coarsen the patch elements            */
/****************************************************************************/

static void  coarse_restrict(RC_LIST_EL *list, int n_el);

static void  coarse_patch(RC_LIST_EL *coarse_list, int n_neigh, int bound)
{
  EL       *el = coarse_list->el;
  REAL     *new_coord = el->new_coord;
  int      i, node;
  DOF      *dof;
  int      *n_dof = coarse_mesh->n_dof;

  if(!coarse_mesh->preserve_coarse_dofs) {
    if (n_dof[EDGE])
      {
/****************************************************************************/
/*  get dof for coarsening edge                                             */
/****************************************************************************/
	node = coarse_mesh->node[EDGE];
	if (!(dof = coarse_list->el->dof[node]))
	  dof = get_dof(coarse_mesh, EDGE);
      }
    else
      dof = nil;

    if (n_dof[EDGE]  ||  n_dof[FACE]  ||  n_dof[CENTER])
      {
	for (i = 0; i < n_neigh; i++)
	  add_dof_parent(coarse_list+i, dof);
      }
  }

  if (call_coarse_restrict)
  {
/****************************************************************************/
/*  restrict dof vectors to the parents on the patch                        */
/****************************************************************************/

    coarse_restrict(coarse_list, n_neigh);
  }

/****************************************************************************/
/*  and now start to coarsen the patch:                                     */
/****************************************************************************/
/*  remove dof's of the coarsening edge                                     */
/****************************************************************************/

  free_dof(el->child[0]->dof[3], coarse_mesh, VERTEX);
  coarse_mesh->n_vertices--;

  if (coarse_mesh->n_dof[EDGE])
  {
    node = coarse_mesh->node[EDGE]+2;
    free_dof(el->child[0]->dof[node], coarse_mesh, EDGE);
    free_dof(el->child[1]->dof[node], coarse_mesh, EDGE);
  }

/*--------------------------------------------------------------------------*/
/*--- and now coarse single elements                                     ---*/
/*--------------------------------------------------------------------------*/

  for (i = 0; i < n_neigh; i++)
  {
    coarse_element(coarse_list+i);
    coarse_list[i].el->new_coord = nil;
  }

/*--------------------------------------------------------------------------*/
/*---  if there is a projected boundary node, free memory                ---*/
/*--------------------------------------------------------------------------*/
  if (new_coord) free_real_d(coarse_mesh, new_coord);

/****************************************************************************/
/*  if the coarsening edge is an interior edge there are  n_neigh + 1 edges */
/*  and 2*n_neigh + 1 faces removed; if it is a boundary edge it is one     */
/*  more edge and one more face                                             */
/****************************************************************************/

  if (bound)
  {
    coarse_mesh->n_edges -= n_neigh + 2;
    coarse_mesh->n_faces -= 2*n_neigh + 1;
  }
  else
  {
    coarse_mesh->n_edges -= n_neigh + 1;
    coarse_mesh->n_faces -= 2*n_neigh;
  }

  return;
}

/****************************************************************************/
/* include common traversal routines for 2d and 3d coarsening               */
/****************************************************************************/

#include "coarsen_common.c"
