/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     element.c                                                      */
/*                                                                          */
/*                                                                          */
/* description:  routines on elements that depend on the dimension in 3d    */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/*--------------------------------------------------------------------------*/
/*  world_to_coord():  return -1 if inside, otherwise index of lambda<0     */
/*--------------------------------------------------------------------------*/

int world_to_coord(const EL_INFO *el_info, const REAL *xy, REAL lambda[DIM+1])
{
  FUNCNAME("world_to_coord");
  REAL  edge[DIM][DIM_OF_WORLD], x[DIM_OF_WORLD];
  REAL  x0, det, det0, det1, det2, adet, lmin;
  int   i, j, k;

#if (DIM != DIM_OF_WORLD)
  ERROR_EXIT("not yet for DIM != DIM_OF_WORLD\n");
#endif

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

/*  wir haben das gleichungssystem zu loesen: */
/*       ( q1x q2x q3x)  (lambda1)     (qx)      */
/*       ( q1y q2y q3y)  (lambda2)  =  (qy)      */
/*       ( q1z q2z q3z)  (lambda3)     (qz)      */
/*      mit qi=pi-p3, q=xy-p3                 */

  for (j=0; j<DIM_OF_WORLD; j++) {
    x0 = el_info->coord[DIM][j];
    x[j] = xy[j] - x0;
    for (i=0; i<DIM; i++)
      edge[i][j] = el_info->coord[i][j] - x0;
  }

  det =  edge[0][0] * edge[1][1] * edge[2][2]
       + edge[0][1] * edge[1][2] * edge[2][0]
       + edge[0][2] * edge[1][0] * edge[2][1]
       - edge[0][2] * edge[1][1] * edge[2][0]
       - edge[0][0] * edge[1][2] * edge[2][1]
       - edge[0][1] * edge[1][0] * edge[2][2];
  det0 =       x[0] * edge[1][1] * edge[2][2]
       +       x[1] * edge[1][2] * edge[2][0]
       +       x[2] * edge[1][0] * edge[2][1]
       -       x[2] * edge[1][1] * edge[2][0]
       -       x[0] * edge[1][2] * edge[2][1]
       -       x[1] * edge[1][0] * edge[2][2];
  det1 = edge[0][0] *       x[1] * edge[2][2]
       + edge[0][1] *       x[2] * edge[2][0]
       + edge[0][2] *       x[0] * edge[2][1]
       - edge[0][2] *       x[1] * edge[2][0]
       - edge[0][0] *       x[2] * edge[2][1]
       - edge[0][1] *       x[0] * edge[2][2];
  det2 = edge[0][0] * edge[1][1] *       x[2]
       + edge[0][1] * edge[1][2] *       x[0]
       + edge[0][2] * edge[1][0] *       x[1]
       - edge[0][2] * edge[1][1] *       x[0]
       - edge[0][0] * edge[1][2] *       x[1]
       - edge[0][1] * edge[1][0] *       x[2];

  adet = ABS(det);

  if (adet < 1.E-20) {
    ERROR_EXIT("det = %le; abort\n", det);
    return(-2);
  }

  lambda[0] = det0 / det;
  lambda[1] = det1 / det;
  lambda[2] = det2 / det;
  lambda[3] = 1.0 - lambda[0] - lambda[1] - lambda[2];

  k = -1;
  lmin = 0.0;
  j = 0;
  for (i = 0; i <= DIM; i++) {
    if ((lambda[i]*adet) < -1.E-15) {
      if (lambda[i] < lmin) {
	k = i;
	lmin = lambda[i];
      }
      j++;
    }
  }

  return(k);
}

/*--------------------------------------------------------------------------*/
/*  transform local coordintes l to world coordinates; if w is non nil      */
/*  store them at w otherwise return a pointer to some local static         */
/*  area containing world coordintes                                        */
/*--------------------------------------------------------------------------*/

const REAL *coord_to_world(const EL_INFO *el_info, const REAL *l, REAL_D w)
{
  FUNCNAME("coord_to_world");
  static REAL world[DIM_OF_WORLD];
  REAL        c, *ret;
  const REAL  *v;
  int         i, j;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  ret = w ? w : world;

  v = el_info->coord[0];
  c = l[0];
  for (j = 0; j < DIM_OF_WORLD; j++)
    ret[j] = c*v[j];
  
  for (i = 1; i < N_VERTICES; i++)
  {
    v = el_info->coord[i];
    c = l[i];
    for (j = 0; j < DIM_OF_WORLD; j++)
      ret[j] += c*v[j];
  }
  return((const REAL *) ret);
}

/*--------------------------------------------------------------------------*/
/*  compute volume of an element                                            */
/*--------------------------------------------------------------------------*/

REAL el_det(const EL_INFO *el_info)
{
  FUNCNAME("el_det");
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], e3[DIM_OF_WORLD], det;
  int         i;
  const REAL  *v0;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
    e3[i] = el_info->coord[3][i] - v0[i];
  }

#if DIM_OF_WORLD==3
  det =   e1[0] * (e2[1]*e3[2] - e2[2]*e3[1])
        - e1[1] * (e2[0]*e3[2] - e2[2]*e3[0])
        + e1[2] * (e2[0]*e3[1] - e2[1]*e3[0]);
  det = ABS(det);
#else
  TEST_EXIT(0)("not yet for DIM_OF_WORLD = %d\n",DIM_OF_WORLD);
#endif

  return(det);
}

#define VOL_THAT  1.0/6.0

REAL el_volume(const EL_INFO *el_info)
{
  return(VOL_THAT*el_det(el_info));
}

/*--------------------------------------------------------------------------*/
/*  compute gradients of basis functions on element                         */
/*  - returns abs(determinante)                                             */
/*--------------------------------------------------------------------------*/

REAL el_grd_lambda(const EL_INFO *el_info, REAL grd_lam[DIM+1][DIM_OF_WORLD])
{
  FUNCNAME("el_grd_lambda");
  int         i, j;
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], e3[DIM_OF_WORLD];
  const REAL  *v0;
  REAL        det, adet;
  REAL        a11, a12, a13, a21, a22, a23, a31, a32, a33;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++) {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
    e3[i] = el_info->coord[3][i] - v0[i];
  }

#if DIM_OF_WORLD == 3
  det =   e1[0] * (e2[1]*e3[2] - e2[2]*e3[1])
        - e1[1] * (e2[0]*e3[2] - e2[2]*e3[0])
        + e1[2] * (e2[0]*e3[1] - e2[1]*e3[0]);
  adet = ABS(det);
  if (adet < 1.0E-25)
  {
    MSG("abs(det) = %lf\n",adet);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	grd_lam[i][j] = 0.0;
  }
  else 
  {
    det = 1.0 / det;
    a11 = (e2[1]*e3[2] - e2[2]*e3[1]) * det;    /* (a_ij) = A^{-T} */
    a12 = (e2[2]*e3[0] - e2[0]*e3[2]) * det;
    a13 = (e2[0]*e3[1] - e2[1]*e3[0]) * det;
    a21 = (e1[2]*e3[1] - e1[1]*e3[2]) * det;
    a22 = (e1[0]*e3[2] - e1[2]*e3[0]) * det;
    a23 = (e1[1]*e3[0] - e1[0]*e3[1]) * det;
    a31 = (e1[1]*e2[2] - e1[2]*e2[1]) * det;
    a32 = (e1[2]*e2[0] - e1[0]*e2[2]) * det;
    a33 = (e1[0]*e2[1] - e1[1]*e2[0]) * det;

    grd_lam[1][0] = a11;
    grd_lam[1][1] = a12;
    grd_lam[1][2] = a13;
    grd_lam[2][0] = a21;
    grd_lam[2][1] = a22;
    grd_lam[2][2] = a23;
    grd_lam[3][0] = a31;
    grd_lam[3][1] = a32;
    grd_lam[3][2] = a33;

    grd_lam[0][0] = -grd_lam[1][0] -grd_lam[2][0] -grd_lam[3][0];
    grd_lam[0][1] = -grd_lam[1][1] -grd_lam[2][1] -grd_lam[3][1];
    grd_lam[0][2] = -grd_lam[1][2] -grd_lam[2][2] -grd_lam[3][2];
  }


#else
  ERROR_EXIT("not yet for DIM_OF_WORLD = %d", DIM_OF_WORLD);
#endif

  return(adet);
}

#if DIM_OF_WORLD == 3
/*--------------------------------------------------------------------------*/
/*  vector product in R^3                                                   */
/*--------------------------------------------------------------------------*/

void vector_product(const REAL_D x, const REAL_D y, REAL_D z)
{
  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  return;
}
#endif

/*--------------------------------------------------------------------------*/
/*  calculate a facenormal to triangular face side of a tetrahedron with    */
/*  coordinates coord; return the absulute value of the determinante from   */
/*  the transformation to the reference element                             */
/*--------------------------------------------------------------------------*/

REAL get_face_normal(const EL_INFO *el_info, int i0, REAL *normal)
{
  FUNCNAME("get_face_normal");
#if DIM_OF_WORLD == 3
  static int   ind[7] = {0, 1, 2, 3, 0, 1, 2};
  REAL         det, scale;
  int          n, i1 = ind[i0+1], i2 = ind[i0+2], i3 = ind[i0+3];
  const REAL_D *coord = el_info->coord;

  REAL_D e0, e1, e2;
  TEST_FLAG(FILL_COORDS, el_info);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    e0[n] = coord[i2][n] - coord[i1][n];
    e1[n] = coord[i3][n] - coord[i1][n];
    e2[n] = coord[i0][n] - coord[i1][n];
  }

  vector_product(e0, e1, normal);

  det = NORM_DOW(normal);
  TEST_EXIT(det > 1.e-30)("det = 0 on face %d\n", i0);

  scale = SCP_DOW(e2,normal) < 0.0 ? 1.0/det : -1.0/det;
  SCAL_DOW(scale, normal);

  return(det);
#else
  MSG("not implemented for DIM_OF_WORLD = %d in 3d\n", DIM_OF_WORLD);
  return(-1.0);
#endif
}

/*--------------------------------------------------------------------------*/
/* orient the vertices of edges/faces                                       */
/* used by estimator for the jumps => same quadrature nodes from both sides!*/
/*--------------------------------------------------------------------------*/

const int *sort_face_indices(const EL *el, int face, int *vec)
{
  FUNCNAME("sort_face_indices");
  static int  sorted[4][7][3] = 
      {{{0,0,0}, {1,3,2}, {2,1,3}, {1,2,3}, {3,2,1}, {3,1,2}, {2,3,1}},
       {{0,0,0}, {0,3,2}, {2,0,3}, {0,2,3}, {3,2,0}, {3,0,2}, {2,3,0}},
       {{0,0,0}, {0,3,1}, {1,0,3}, {0,1,3}, {3,1,0}, {3,0,1}, {1,3,0}},
       {{0,0,0}, {0,2,1}, {1,0,2}, {0,1,2}, {2,1,0}, {2,0,1}, {1,2,0}}};
  static int  vertex_of_face[4][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};
  int    i, no, *val  = nil;
  int    *vof = vertex_of_face[face];
  DOF    **dof = el->dof;

  no = 0;
  if (dof[vof[0]][0] < dof[vof[1]][0])
    no++;
  if (dof[vof[1]][0] < dof[vof[2]][0])
    no += 2;
  if (dof[vof[2]][0] < dof[vof[0]][0])
    no += 4;
  
  if (no >= 1  &&  no <= 6)
  {
    if (vec)
    {
      val = vec;
      for (i = 0; i < 3; i++)
	val[i] = sorted[face][no][i];
    }
    else
      val = sorted[face][no];
  }
  else
  {
    MSG("can not sort face indices of element %d at face %d\n", 
	INDEX(el), face);
  }
 
  return((const int *) val);
}
