/****************************************************************************/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     HB_precon.c                                                    */
/*                                                                          */
/* description:  hierachical basis preconditioning, including higher        */
/*               order                                                      */
/*               and also the BPX preconditioner                            */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*---8<---------------------------------------------------------------------*/
/*---  to do: for time dependent problems, a modification of constants   ---*/
/*---         in BPX is necessary for optimal convergence...             ---*/
/*---                                                                    ---*/
/*---  to do: something is wrong in BPX for p > 2 ???                    ---*/
/*---  BPX does not always converge for p>2 (but does for high refinement)--*/
/*--------------------------------------------------------------------->8---*/

#include "alberta.h"

typedef struct
{ 
  const FE_SPACE      *fe_space;
  const DOF_SCHAR_VEC *bound_dv;
  int                 use_get_bound;

  int                 info;
  int                 high_degree;
  int                 mg_levels;
  int                 size;
  U_CHAR              *dof_level;
  U_CHAR              *local_dof;
  DOF                 (*dof_parent)[DIM+1];
  DOF                 *sort_dof, *sort_dof_invers;
  int                 *dofs_per_level;
  const S_CHAR        *bound;
  REAL                (*ipol)[DIM+1];      /* for high_degree interpolation */

  /* BPX section: */
  REAL                *g;
  REAL                diam;
} HB_DATA;


static void free_HB_BPX_precon(void *vdata, int BPX);

/****************************************************************************/

static int       n0_vert=0, max_level=0, max_dof_level=0;
static const int *n_dof=nil, *n0_dof=nil, *node=nil;
static int       *local_dof_sort=nil;
static HB_DATA   *hb_data=nil;

static void max_level_fct(const EL_INFO *el_info)
{
  int dof, dof0, dof1, level = (int)(el_info->level);
  EL  *el   = el_info->el;

  max_level = MAX(level, max_level);

  if (!(IS_LEAF_EL(el))) {
    dof   = el->child[0]->dof[N_VERTICES-1][n0_vert];
    dof0  = hb_data->dof_parent[dof][0] = el->dof[0][n0_vert];
    dof1  = hb_data->dof_parent[dof][1] = el->dof[1][n0_vert];
    level = 1 + MAX(hb_data->dof_level[dof0], hb_data->dof_level[dof1]);
    hb_data->dof_level[dof] = level; 
    max_dof_level = MAX(level, max_dof_level);
  }
}


/* high_degree_fct(): set level of all non-vertex dofs to highest level,  */
/*                    and save parent data                                */

static void high_degree_fct(const EL_INFO *el_info)
{
  FUNCNAME("high_degree_fct");
#if DIM > 1
  int  i;
#endif
  int  j, k, m, n, nod0, n0, dof;
  EL   *el   = el_info->el;
  DOF  vertex_dof[N_VERTICES];
  
  if (IS_LEAF_EL(el)) {
    for (k=0; k<N_VERTICES; k++)
      vertex_dof[k] = el->dof[k][n0_vert];
    m = N_VERTICES;

#if DIM > 1    
    if ((n = n_dof[EDGE]) > 0) {
      nod0 = node[EDGE];
      n0 = n0_dof[EDGE];
      for (i=0; i<N_EDGES; i++)
	for (j=0; j<n; j++) {
	  dof = el->dof[nod0+i][n0+j];
	  if (!hb_data->local_dof[dof]) {
/*  	    MSG("edge dof   %2d: level %d\n", dof, max_dof_level); */
	    hb_data->dof_level[dof] = max_dof_level;
	    for (k=0; k<N_VERTICES; k++)
	      hb_data->dof_parent[dof][k] = vertex_dof[k];
	    hb_data->local_dof[dof] = local_dof_sort[m];
	  }
	  m++;
	}
      
    }
#endif

#if DIM==3
    if ((n = n_dof[FACE]) > 0) {
      nod0 = node[FACE];
      n0 = n0_dof[FACE];
      for (i=0; i<N_FACES; i++)
	for (j=0; j<n; j++) {
	  dof = el->dof[nod0+i][n0+j];
	  if (!hb_data->local_dof[dof]) {
/*  	    MSG("face dof   %2d: level %d\n", dof, max_dof_level); */
	    hb_data->dof_level[dof] = max_dof_level;
	    for (k=0; k<N_VERTICES; k++)
	      hb_data->dof_parent[dof][k] = vertex_dof[k];
	    hb_data->local_dof[dof] = local_dof_sort[m];
	  }
	  m++;
	}
    }
#endif
    if ((n = n_dof[CENTER]) > 0) {
      nod0 = node[CENTER];
      n0 = n0_dof[CENTER];
      for (j=0; j<n; j++) {
	  dof = el->dof[nod0][n0+j];
/*  	  MSG("center dof %2d: level %d\n", dof, max_dof_level); */
	  hb_data->dof_level[dof] = max_dof_level;
	  for (k=0; k<N_VERTICES; k++)
	    hb_data->dof_parent[dof][k] = vertex_dof[k];
	  hb_data->local_dof[dof] = local_dof_sort[m++];
	}
    }
    TEST_EXIT(m == hb_data->fe_space->bas_fcts->n_bas_fcts)
      ("m <> n_bas_fcts: %d  %d\n", m, hb_data->fe_space->bas_fcts->n_bas_fcts);
  }
  else {
    ERROR_EXIT("non-leaf element???\n");
  }
  return;
}


/****************************************************************************/
#if DIM==1
static DOF *el_hat_dofs[N_VERTICES+1];
#endif
#if DIM==2
static DOF *el_hat_dofs[N_VERTICES+N_EDGES+1];
#endif
#if DIM==3
static DOF *el_hat_dofs[N_VERTICES+N_EDGES+N_FACES+1];
#endif

static MESH dummy_mesh = {0};

static EL el_hat =
  { 
    { nil, nil},  /* children */
    el_hat_dofs,
    0    /* ... */
  };

static EL_INFO el_hat_info = 
  { &dummy_mesh,
#if DIM==1
    { {0.0}, {1.0} },
#endif
#if DIM==2
    { {1.0, 0.0}, {0.0, 1.0}, {0.0, 0.0} },
#endif
#if DIM==3
    { {1.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0.0, 0.0, 1.0}, {0.0, 0.0, 0.0} },
#endif
    nil,
    nil, nil,
    FILL_COORDS,
    { 0 }  /*---  and the rest is initialized with 0s  ---*/
  };

#if DIM==1
static REAL lambda_0(const REAL_D x) { return(1.0-x[0]);}
static REAL lambda_1(const REAL_D x) { return(x[0]);}
static REAL ((*lambda[N_VERTICES])(const REAL_D))
  = {lambda_0,lambda_1};
#endif

#if DIM==2
static REAL lambda_0(const REAL_D x) { return(x[0]);}
static REAL lambda_1(const REAL_D x) { return(x[1]);}
static REAL lambda_2(const REAL_D x) { return(1.0-x[0]-x[1]);}
static REAL ((*lambda[N_VERTICES])(const REAL_D))
  = {lambda_0,lambda_1,lambda_2};
#endif

#if DIM==3
static REAL lambda_0(const REAL_D x) { return(x[0]);}
static REAL lambda_1(const REAL_D x) { return(x[1]);}
static REAL lambda_2(const REAL_D x) { return(x[2]);}
static REAL lambda_3(const REAL_D x) { return(1.0-x[0]-x[1]-x[2]);}
static REAL ((*lambda[N_VERTICES])(const REAL_D))
  = {lambda_0,lambda_1,lambda_2,lambda_3};
#endif

/****************************************************************************/
/*  init_HB_precon(matrix, boundary, info);    initialize HB_DATA structure */
/*  fills    (void *)(HB_DATA *), if ok,                                    */
/*  returns   nil, if anything goes wrong.                                  */
/****************************************************************************/

static int init_HB_BPX_precon(void *precon_data, int BPX)
{
  FUNCNAME("init_HB_BPX_precon");
  HB_DATA         *data = precon_data;
  const FE_SPACE  *fe_space;
  const DOF_ADMIN *admin;
  int             i, j, k, m, size, info;
  int             *tmp_per_level = nil;
  int             n_bas_fcts;
  const REAL      *one_ipol;
  DOF             *tmp_dofs, *tmp_dof_ptr;
  const DOF       *dof_indices;

  if (!data) 
  {
      ERROR("no precon_data\n");
      return false;
  }
  if (!(data->fe_space)) 
  {
      ERROR("no precon_data->fe_space\n");
      return false;
  }
  
  info     = data->info;
  fe_space = data->fe_space;
  if (!fe_space || !(fe_space->admin) || !(fe_space->bas_fcts))
  {
    MSG("no fe_space or admin or bas_fcts.\n");
    return false;
  }

  if (fe_space->bas_fcts->n_dof[VERTEX] != 1)
  {
    MSG("sorry, only for FE spaces with n_dof[VERTEX]==1.\n");
    return false;
  }


/*  When init() is called twice, free all vectors before reallocating */
  if (data->dof_level) {
      free_HB_BPX_precon(data, BPX);
  }



  admin = fe_space->admin;
  size  = admin->size_used;

  n_bas_fcts = fe_space->bas_fcts->n_bas_fcts;
  data->high_degree = (n_bas_fcts > N_VERTICES);
  if (data->high_degree)
  {
    INFO(info, 1)("use high degree version\n");
  }

  data->mg_levels = 0;
  data->dofs_per_level = nil;

  if (data->use_get_bound)
  {
    DOF_SCHAR_VEC  dsv = {nil, nil, "HB/BPX bound", 0, nil, nil, nil};

    dsv.fe_space = data->fe_space;
    dsv.size     = size;
    dsv.vec      = MEM_ALLOC(size, S_CHAR);

    dirichlet_bound(nil, nil, nil, &dsv);
    data->bound = dsv.vec;
  }
  else if (data->bound_dv)
  {
    data->bound = data->bound_dv->vec;
  }
  else
  {
    data->bound = nil;
  }

  data->ipol = nil;
  
  data->dof_level       = MEM_ALLOC(2*size, U_CHAR);
  data->local_dof       = data->dof_level       + size;
  
  data->sort_dof        = MEM_ALLOC((DIM+3)*size, DOF);
  data->sort_dof_invers = data->sort_dof        + size;
  data->dof_parent      = (void *)(data->sort_dof_invers + size);
  data->size            = size;

  FOR_ALL_DOFS(admin,
	       data->dof_level[dof] = 0;
	       for (j=0; j <= DIM; j++) data->dof_parent[dof][j] = -1;
	       data->local_dof[dof] = 0;
	       );

  n0_vert = admin->n0_dof[VERTEX];
  max_level = max_dof_level = 0;
  hb_data   = data;
  mesh_traverse(fe_space->mesh, -1, CALL_EVERY_EL_PREORDER, max_level_fct);
  data->mg_levels = (max_level + DIM-1) / DIM + 1;

  TEST_EXIT(data->mg_levels == (max_dof_level+1))
    ("mg_levels %d != max_dof_level %d + 1\n", data->mg_levels, max_dof_level);


  if (data->high_degree) {     /* add level for fine-grid non-vertex DOFs */
    data->mg_levels++;
    data->ipol = (void *)(MEM_ALLOC((DIM+1)*n_bas_fcts, REAL));

    local_dof_sort = MEM_ALLOC(n_bas_fcts, int);
    for (j=0; j<n_bas_fcts; j++) local_dof_sort[j] = j;  /* ???? */
    
    tmp_dof_ptr = tmp_dofs = MEM_ALLOC(fe_space->mesh->n_dof_el, DOF);
    m = 0;
    for (i=0; i<N_VERTICES; i++) {
      el_hat_dofs[fe_space->mesh->node[VERTEX]+i] = tmp_dof_ptr;
      for (j=0; j<admin->n_dof[VERTEX]; j++)
	tmp_dof_ptr[admin->n0_dof[VERTEX]+j] = m++;
      tmp_dof_ptr += fe_space->mesh->n_dof[VERTEX];
    }
#if DIM > 1
    if (fe_space->mesh->n_dof[EDGE])
      for (i=0; i<N_EDGES; i++) {
	el_hat_dofs[fe_space->mesh->node[EDGE]+i] = tmp_dof_ptr;
	for (j=0; j<admin->n_dof[EDGE]; j++)
	  tmp_dof_ptr[admin->n0_dof[EDGE]+j] = m++;
	tmp_dof_ptr += fe_space->mesh->n_dof[EDGE];
      }
#endif

#if DIM==3
    if (fe_space->mesh->n_dof[FACE])
      for (i=0; i<N_FACES; i++) {
	el_hat_dofs[fe_space->mesh->node[FACE]+i] = tmp_dof_ptr;
	for (j=0; j<admin->n_dof[FACE]; j++)
	  tmp_dof_ptr[admin->n0_dof[FACE]+j] = m++;
	tmp_dof_ptr += fe_space->mesh->n_dof[FACE];
      }
#endif

    if (fe_space->mesh->n_dof[CENTER]) {
      el_hat_dofs[fe_space->mesh->node[CENTER]] = tmp_dof_ptr;
      for (j=0; j<admin->n_dof[CENTER]; j++)
	tmp_dof_ptr[admin->n0_dof[CENTER]+j] = m++;
    }
    TEST_EXIT(m==n_bas_fcts)("m != n_bas_fcts: %d %d\n", m, n_bas_fcts);
    
    dof_indices = fe_space->bas_fcts->get_dof_indices(&el_hat, admin, nil);
    for (i = 0; i<fe_space->bas_fcts->n_bas_fcts; i++)
      local_dof_sort[dof_indices[i]] = i;

#if 0
    print_int_vec("dof_indices   ", dof_indices, n_bas_fcts);
    print_int_vec("local_dof_sort", local_dof_sort, n_bas_fcts);
#endif

    MEM_FREE(tmp_dofs, fe_space->mesh->n_dof_el, DOF);



    for (i=0; i<N_VERTICES; i++) {
      one_ipol =
	fe_space->bas_fcts->interpol(&el_hat_info, 0, nil, lambda[i], nil);
      for (j=0; j<n_bas_fcts; j++)
	data->ipol[j][i] = one_ipol[j];
    }

   
    max_dof_level++;
    n0_dof = admin->n0_dof;
    n_dof  = admin->n_dof;
    node   = fe_space->mesh->node;
    
    mesh_traverse(fe_space->mesh, -1, CALL_LEAF_EL, high_degree_fct);    

    MEM_FREE(local_dof_sort, n_bas_fcts, int);

    if (info > 3) 
    {
      for (i=0; i<n_bas_fcts; i++)
      {
	MSG("ipol[%2d]", i); PRINT_REAL_VEC("", data->ipol[i], N_VERTICES);
      }
    }
    
#if 0
    FOR_ALL_DOFS(admin,
		 if (data->local_dof[dof])
		   MSG("dof %3d: local_dof=%2d\n", dof, data->local_dof[dof]);
		 );
#endif
  }
  
  if (data->mg_levels < 2) {
      free_HB_BPX_precon(data, BPX);
      return true;
  }
 
  data->dofs_per_level = MEM_ALLOC(data->mg_levels, int);
  tmp_per_level  = MEM_ALLOC(data->mg_levels, int);
  
  for (i = 0; i < data->mg_levels; i++) data->dofs_per_level[i] = 0;
  FOR_ALL_DOFS(admin, data->dofs_per_level[data->dof_level[dof]]++);
  if (info > 3) {
    MSG("dofs_per_level:");
    for (i = 0; i < data->mg_levels; i++)
      print_msg(" %d", data->dofs_per_level[i]);
    print_msg("\n");
  }
  for (i = 1; i < data->mg_levels; i++) {
    tmp_per_level[i]   = data->dofs_per_level[i-1];
    data->dofs_per_level[i] += data->dofs_per_level[i-1];
  }
  tmp_per_level[0] = 0;      /* pointers for filling the sort vectors */

  if (info > 3) {
    MSG("dofs_per_level accumulated:");
    for (i = 0; i < data->mg_levels; i++)
      print_msg(" %d", data->dofs_per_level[i]);
    print_msg("\n");
  }

#if 0
  for (i = 0; i < data->dofs_per_level[data->mg_levels-1]; i++)
    MSG("dof_parent[%3d] = (%3d,%3d,%3d), lev=%2d (%2d,%2d,%3d)\n",
	i, 
	data->dof_parent[i][0], data->dof_parent[i][1], data->dof_parent[i][2],
	data->dof_level[i], 
	data->dof_level[data->dof_parent[i][0]],
	data->dof_level[data->dof_parent[i][1]],
	data->dof_level[data->dof_parent[i][2]]);
#endif


/* build sort_dof[] and sort_dof_invers[] vectors                           */

  FOR_ALL_DOFS(admin,
	       j = data->dof_level[dof];
	       k = tmp_per_level[j]++;
	       data->sort_dof[k] = dof;
	       data->sort_dof_invers[dof] = k;
    );

#if 0
  for (i = 0; i < data->dofs_per_level[data->mg_levels-1]; i++)
  {
    j = data->sort_dof[i];
    MSG("sort[%3d]: dof=%3d, lev=%2d; invers[%3d]=%3d\n",
	i, j, data->dof_level[j], j, data->sort_dof_invers[j]);
  }
  /* WAIT; */
#endif

  MEM_FREE(tmp_per_level, data->mg_levels, int);

  if (BPX) 
  {
    data->g = MEM_CALLOC(DIM_OF_WORLD*data->size, REAL);
    data->diam = fe_space->mesh->diam[0];
    for (i=1; i<DIM_OF_WORLD; i++)
      data->diam = MAX(data->diam, fe_space->mesh->diam[i]);
  }
  else
  {
    data->g = nil;
    data->diam = 0;
  }

  return true;
}


/****************************************************************************/
/****************************************************************************/

static void free_HB_BPX_precon(void *vdata, int BPX)
{
  FUNCNAME("free_HB_BPX_precon");
  HB_DATA *data = vdata;

  if (!data) 
  {
    MSG("no data ???\n");
    return;
  }

  if (BPX) 
  {
/*     TEST(data->g)("no g entry in BPX data???\n"); */
  }
  else 
  {
/*     TEST(!(data->g))("g entry in HB data???\n"); */
  }
  if (data->g)
      MEM_FREE(data->g, DIM_OF_WORLD*data->size, REAL);
  data->g = nil;
  
  if (data->dofs_per_level)
      MEM_FREE(data->dofs_per_level, data->mg_levels, int);
  data->dofs_per_level = nil;

  if (data->high_degree && data->ipol) 
  {
    MEM_FREE(data->ipol, (DIM+1)*data->fe_space->bas_fcts->n_bas_fcts, REAL);
    data->ipol = nil;
  }

  if (data->sort_dof)
      MEM_FREE(data->sort_dof, (DIM+3)*data->size, DOF);
  data->sort_dof = nil;
  if (data->dof_level)
      MEM_FREE(data->dof_level, 2*data->size, U_CHAR);
  data->dof_level = nil;

  if (data->bound)
      MEM_FREE(data->bound, data->size, S_CHAR);
  data->bound = nil;

  data->size = 0;

  return;
}


static void exit_HB_BPX_precon(void *vdata, int BPX)
{
    if (vdata) {
	free_HB_BPX_precon(vdata, BPX);
	MEM_FREE(vdata, 1, HB_DATA);
    }
    return;
}

/****************************************************************************/
int init_HB_precon(void *precon_data)
{
  return init_HB_BPX_precon(precon_data, 0);
}
void exit_HB_precon(void *precon_data)
{
  exit_HB_BPX_precon(precon_data, 0);
}
/****************************************************************************/

void HB_precon_s(void *vdata, int n, REAL *r)
{
  FUNCNAME("HB_precon_s");
  int     i, idof, level, last, jdof, k, level1;
  HB_DATA *data;

  data = (HB_DATA *)vdata;
  if (!data) {
    MSG("no data ???\n");
    return;
  }

  if (data->mg_levels < 2)                                /* nothing to do */
  {
    return;
  }

  if (n > data->size) {
    MSG("n > data->size ???\n");
    return;
  }

/* transposed basis transformation (run over refined levels only)          */

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];
    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      jdof = data->local_dof[i];
      if (data->bound) {
	for (k=0; k<N_VERTICES; k++) {
	  if (data->bound[data->dof_parent[idof][k]] <= INTERIOR)
	    r[data->dof_parent[idof][k]] += data->ipol[jdof][k] * r[idof];
	}
      }
      else {
	for (k=0; k<N_VERTICES; k++) {
	  r[data->dof_parent[idof][k]] += data->ipol[jdof][k] * r[idof];
	}
      }
    }
    level1 = data->mg_levels - 2;
  } 
  else {
    level1 = data->mg_levels - 1;
  }
  


  for (level = level1; level > 0; level--)
  {
    last = data->dofs_per_level[level];

    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (data->bound) {
	if (data->bound[data->dof_parent[idof][0]] <= INTERIOR)
	  r[data->dof_parent[idof][0]] += 0.5 * r[idof];
	if (data->bound[data->dof_parent[idof][1]] <= INTERIOR)
	  r[data->dof_parent[idof][1]] += 0.5 * r[idof];
      }
      else {
	r[data->dof_parent[idof][0]] += 0.5 * r[idof];
	r[data->dof_parent[idof][1]] += 0.5 * r[idof];
      }
    }
  }

/* basis transformation (run over refined levels only) */

  for (level = 1; level <= level1; level++)
  {
    last = data->dofs_per_level[level];

    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	r[idof] += 0.5 * (r[data->dof_parent[idof][0]]
			  + r[data->dof_parent[idof][1]]);
      }
    }
  }

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];
    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	jdof = data->local_dof[i];
	for (k = 0; k < N_VERTICES; k++)
	  r[idof] += data->ipol[jdof][k] * r[data->dof_parent[idof][k]];
      }
    }
  }

  return;
}

/****************************************************************************/

void HB_precon_d(void *vdata, int n, REAL *r)
{
  FUNCNAME("HB_precon_d");
  int     i,j, idof, level, last, jdof, k, level1;
  HB_DATA *data;
  REAL_D  *rd = (REAL_D *)r;

  data = (HB_DATA *)vdata;
  if (!data) {
    MSG("no data ???\n");
    return;
  }

  if (data->mg_levels < 2)                                /* nothing to do */
  {
    return;
  }

  if (n > DIM_OF_WORLD*data->size) {
    MSG("n > data->DIM_OF_WORLD*size ???\n");
    return;
  }

/* transposed basis transformation (run over refined levels only)          */

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];
    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      jdof = data->local_dof[i];
      if (data->bound) {
	for (k=0; k<N_VERTICES; k++) {
	  if (data->bound[data->dof_parent[idof][k]] <= INTERIOR)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      rd[data->dof_parent[idof][k]][j]
		+= data->ipol[jdof][k] * rd[idof][j];
	}
      }
      else {
	for (k=0; k<N_VERTICES; k++) {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    rd[data->dof_parent[idof][k]][j]
	      += data->ipol[jdof][k] * rd[idof][j];
	}
      }
    }
    level1 = data->mg_levels - 2;
  } 
  else {
    level1 = data->mg_levels - 1;
  }
  

  for (level = level1; level > 0; level--)
  {
    last = data->dofs_per_level[level];

    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (data->bound) {
	if (data->bound[data->dof_parent[idof][0]] <= INTERIOR)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    rd[data->dof_parent[idof][0]][j] += 0.5 * rd[idof][j];
	if (data->bound[data->dof_parent[idof][1]] <= INTERIOR)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    rd[data->dof_parent[idof][1]][j] += 0.5 * rd[idof][j];
      }
      else {
	for (j=0; j<DIM_OF_WORLD; j++) {
	  rd[data->dof_parent[idof][0]][j] += 0.5 * rd[idof][j];
	  rd[data->dof_parent[idof][1]][j] += 0.5 * rd[idof][j];
	}
      }
    }
  }

/* basis transformation (run over refined levels only) */

  for (level = 1; level <= level1; level++)
  {
    last = data->dofs_per_level[level];

    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	for (j=0; j<DIM_OF_WORLD; j++) {
	  rd[idof][j] += 0.5 * (rd[data->dof_parent[idof][0]][j]
			       + rd[data->dof_parent[idof][1]][j]);
	}
      }
    }
  }

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];
    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	jdof = data->local_dof[i];
	for (k = 0; k < N_VERTICES; k++)
	  for (j=0; j<DIM_OF_WORLD; j++) {
	    rd[idof][j]
	      += data->ipol[jdof][k] * rd[data->dof_parent[idof][k]][j];
	  }
      }
    }
  }
  
  return;
}

/****************************************************************************/
/****************************************************************************/

const PRECON *get_HB_precon_s(const FE_SPACE *fe_space,
			      const DOF_SCHAR_VEC *bound, 
			      int use_get_bound, int info)
{
    PRECON  *precon;
    HB_DATA *data;
    
    if (bound && (bound->fe_space != fe_space)) {
	ERROR("different fe spaces ?\n");
	return(nil);
    }

    precon = MEM_CALLOC(1, PRECON);
    data   = MEM_CALLOC(1, HB_DATA);
    
    data->fe_space = fe_space;
    if (!(data->bound_dv = bound))
      data->use_get_bound = use_get_bound;

    data->info     = info;

    precon->precon_data = data;
    precon->init_precon = init_HB_precon;
    precon->precon      = HB_precon_s;
    precon->exit_precon = exit_HB_precon;

    return(precon);
}

const PRECON *get_HB_precon_d(const FE_SPACE *fe_space,
			      const DOF_SCHAR_VEC *bound, 
			      int use_get_bound, int info)
{
    PRECON  *precon;
    HB_DATA *data;
    
    if (bound && (bound->fe_space != fe_space)) {
	ERROR("different fe spaces ?\n");
	return(nil);
    }

    precon = MEM_CALLOC(1, PRECON);
    data   = MEM_CALLOC(1, HB_DATA);
    
    data->fe_space = fe_space;
    if (!(data->bound_dv = bound))
      data->use_get_bound = use_get_bound;

    data->info     = info;

    precon->precon_data = data;
    precon->init_precon = init_HB_precon;
    precon->precon      = HB_precon_d;
    precon->exit_precon = exit_HB_precon;

    return(precon);
}

/****************************************************************************/
/****************************************************************************/
/**   BPX  preconditioner                                                  **/
/****************************************************************************/
/****************************************************************************/

int init_BPX_precon(void *precon_data)
{
  return init_HB_BPX_precon(precon_data, 1);
}

void exit_BPX_precon(void *precon_data)
{
  exit_HB_BPX_precon(precon_data, 1);
}
/****************************************************************************/

void BPX_precon_s(void *vdata, int n, REAL *h)
{
  FUNCNAME("BPX_precon_s");
  int     i, idof, level, first = 0, last, jdof, k, level1, *dof_parent;
  HB_DATA *data;
  REAL    *g, *ipol;

  data = (HB_DATA *)vdata;
  if (!data) {
    MSG("no data ???\n");
    return;
  }

  if (data->mg_levels < 2)                                /* nothing to do */
  {
    return;
  }

  if (n > data->size) {
    MSG("n > data->size ???\n");
    return;
  }

  TEST_EXIT(g = data->g)("no g vec in HB_DATA\n");

  /* copy h to g */
  for (i=0; i<data->size; i++)  g[i] = h[i];
  
  /* Loop over all non-macro levels */

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];

    /* inverse basis transformation on high degree level for h */

    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      dof_parent = data->dof_parent[idof];
      jdof = data->local_dof[i];
      ipol = data->ipol[jdof];
      if (data->bound) {
	for (k=0; k<N_VERTICES; k++) {
	  if (data->bound[data->dof_parent[idof][k]] <= INTERIOR)
	    h[idof] -= ipol[k] * h[dof_parent[k]];
	}
      }
      else
      {
	for (k=0; k<N_VERTICES; k++) {
	  h[idof] -= ipol[k] * h[dof_parent[k]];
	}
      }
    }

    /* transposed basis transformation on high degree level for g */

    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      dof_parent = data->dof_parent[idof];
      jdof = data->local_dof[i];
      ipol = data->ipol[jdof];
      if (data->bound) {
	for (k=0; k<N_VERTICES; k++) {
	  if (data->bound[dof_parent[k]] <= INTERIOR)
	    g[dof_parent[k]] += ipol[k] * g[idof];
	}
      }
      else
      {
	for (k=0; k<N_VERTICES; k++) {
	  g[dof_parent[k]] += ipol[k] * g[idof];
	}
      }
    }

    /* add up to degree1 level */

    last = data->dofs_per_level[data->mg_levels - 2];
    if (data->bound) {
      for (i = 0; i < last; i++) {
	idof = data->sort_dof[i];
	if (data->bound[idof] <= 0)
	  h[idof] += g[idof];
      }
    } else 
    {
      for (i = first; i < last; i++) 
      {
	idof = data->sort_dof[i];
	h[idof] += g[idof];
      }
    }
    level1 = data->mg_levels - 2;
  } else {
    level1 = data->mg_levels - 1;
  }

  for (level = level1; level > 0; level--)
  {
    last = data->dofs_per_level[level];

    /* inverse basis transformation on level k (only) for h */
    for (i = data->dofs_per_level[level-1]; i < last; i++) {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	h[idof] -= 0.5 * (h[data->dof_parent[idof][0]]
			  + h[data->dof_parent[idof][1]]);
      }
    }
    

    /* transposed basis transformation on level k-1 for g */
    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (data->bound) {
	if (data->bound[data->dof_parent[idof][0]] <= INTERIOR)
	  g[data->dof_parent[idof][0]] += 0.5 * g[idof];
	if (data->bound[data->dof_parent[idof][1]] <= INTERIOR)
	  g[data->dof_parent[idof][1]] += 0.5 * g[idof];
      }
      else {
	g[data->dof_parent[idof][0]] += 0.5 * g[idof];
	g[data->dof_parent[idof][1]] += 0.5 * g[idof];
      }
    }
  
    /* add up to level k-1 */


    last = data->dofs_per_level[level-1];
    if (data->bound) {
      for (i = 0; i < last; i++) {
	idof = data->sort_dof[i];
	if (data->bound[idof] <= 0)
	  h[idof] += g[idof];
      }
    } else 
    {
      for (i = first; i < last; i++) 
      {
	idof = data->sort_dof[i];
	h[idof] += g[idof];
      }
    }
  }
  
  /* basis transformation (all refined levels) */

  for (level = 1; level <= level1; level++)
  {
    last = data->dofs_per_level[level];

    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	h[idof] += 0.5 * (h[data->dof_parent[idof][0]]
			  + h[data->dof_parent[idof][1]]);
      }
    }
  }

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];
    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	dof_parent = data->dof_parent[idof];
	jdof = data->local_dof[i];
	ipol = data->ipol[jdof];
	for (k = 0; k < N_VERTICES; k++)
	  h[idof] += ipol[k] * h[dof_parent[k]];
      }
    }
  }

  return;
}

/****************************************************************************/
void BPX_precon_d(void *vdata, int n, REAL *hh)
{
  FUNCNAME("BPX_precon_d");
  int     i, j, k, idof, level, first = 0, last, jdof, level1;
  HB_DATA *data;
  REAL_D  *g;
  REAL_D  *h = (REAL_D *)hh;

  data = (HB_DATA *)vdata;
  if (!data) {
    MSG("no data ???\n");
    return;
  }

  if (data->mg_levels < 2)                                /* nothing to do */
  {
    return;
  }

  if (n > data->size*DIM_OF_WORLD) {
    MSG("n > data->size*DIM_OF_WORLD ???\n");
    return;
  }

  TEST_EXIT(g = (REAL_D *)data->g)("no g vec in HB_DATA\n");

  /* copy h to g */
  for (i=0; i<data->size; i++)
    for (j=0; j<DIM_OF_WORLD; j++)
      g[i][j] = h[i][j];
  
  /* Loop over all non-macro levels */

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];

    /* inverse basis transformation on high degree level for h */

    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      jdof = data->local_dof[i];
      if (data->bound) {
	for (k=0; k<N_VERTICES; k++) {
	  if (data->bound[data->dof_parent[idof][k]] <= INTERIOR)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      h[idof][j]
		-= data->ipol[jdof][k] * h[data->dof_parent[idof][k]][j];
	}
      }
      else {
	for (k=0; k<N_VERTICES; k++) {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    h[idof][j]
	      -= data->ipol[jdof][k] * h[data->dof_parent[idof][k]][j];
	}
      }
    }

    /* transposed basis transformation on high degree level for g */

    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      jdof = data->local_dof[i];
      if (data->bound) {
	for (k=0; k<N_VERTICES; k++) {
	  if (data->bound[data->dof_parent[idof][k]] <= INTERIOR)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      g[data->dof_parent[idof][k]][j] 
		+= data->ipol[jdof][k] * g[idof][j];
	}
      }
      else {
	for (k=0; k<N_VERTICES; k++) {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      g[data->dof_parent[idof][k]][j]
		+= data->ipol[jdof][k] * g[idof][j];
	}
      }
    }

    /* add up to degree1 level */

    last = data->dofs_per_level[data->mg_levels - 2];
    if (data->bound) {
      for (i = 0; i < last; i++) {
	idof = data->sort_dof[i];
	if (data->bound[idof] <= 0)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    h[idof][j] += g[idof][j];
      }
    } else {
      for (i = first; i < last; i++) {
	idof = data->sort_dof[i];
	for (j=0; j<DIM_OF_WORLD; j++)
	  h[idof][j] += g[idof][j];
      }
    }
    
    level1 = data->mg_levels - 2;
  } else {
    level1 = data->mg_levels - 1;
  }

  for (level = level1; level > 0; level--)
  {
    last = data->dofs_per_level[level];

    /* inverse basis transformation on level k (only) for h */
    for (i = data->dofs_per_level[level-1]; i < last; i++) {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	for (j=0; j<DIM_OF_WORLD; j++)
	  h[idof][j] -= 0.5 * (h[data->dof_parent[idof][0]][j]
			       + h[data->dof_parent[idof][1]][j]);
      }
    }
    

    /* transposed basis transformation on level k-1 for g */
    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (data->bound) {
	if (data->bound[data->dof_parent[idof][0]] <= INTERIOR)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    g[data->dof_parent[idof][0]][j] += 0.5 * g[idof][j];
	if (data->bound[data->dof_parent[idof][1]] <= INTERIOR)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    g[data->dof_parent[idof][1]][j] += 0.5 * g[idof][j];
      }
      else {
	for (j=0; j<DIM_OF_WORLD; j++) {
	  g[data->dof_parent[idof][0]][j] += 0.5 * g[idof][j];
	  g[data->dof_parent[idof][1]][j] += 0.5 * g[idof][j];
	}
      }
    }
  
    /* add up to level k-1 */


    last = data->dofs_per_level[level-1];
    if (data->bound) {
      for (i = 0; i < last; i++) {
	idof = data->sort_dof[i];
	if (data->bound[idof] <= 0)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    h[idof][j] += g[idof][j];
      }
    } else {
      for (i = first; i < last; i++) {
	idof = data->sort_dof[i];
	for (j=0; j<DIM_OF_WORLD; j++)
	  h[idof][j] += g[idof][j];
      }
    }

  }
  
  /* basis transformation (all refined levels) */

  for (level = 1; level <= level1; level++)
  {
    last = data->dofs_per_level[level];

    for (i = data->dofs_per_level[level-1]; i < last; i++)
    {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	for (j=0; j<DIM_OF_WORLD; j++)
	  h[idof][j] += 0.5 * (h[data->dof_parent[idof][0]][j]
			       + h[data->dof_parent[idof][1]][j]);
      }
    }
  }

  if (data->high_degree) {
    last = data->dofs_per_level[data->mg_levels - 1];
    for (i = data->dofs_per_level[data->mg_levels - 2]; i < last; i++)  {
      idof = data->sort_dof[i];
      if (!(data->bound && data->bound[idof] >= DIRICHLET)) {
	jdof = data->local_dof[i];
	for (k = 0; k < N_VERTICES; k++)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    h[idof][j]
	      += data->ipol[jdof][k] * h[data->dof_parent[idof][k]][j];
      }
    }
  }
  return;
}

/****************************************************************************/

const PRECON *get_BPX_precon_s(const FE_SPACE *fe_space,
			       const DOF_SCHAR_VEC *bound, 
			       int use_get_bound, int info)
{
    PRECON  *precon;
    HB_DATA *data;
    
    if (bound && (bound->fe_space != fe_space)) {
	ERROR("different fe spaces ?\n");
	return(nil);
    }

    precon = MEM_CALLOC(1, PRECON);
    data   = MEM_CALLOC(1, HB_DATA);
    
    data->fe_space = fe_space;
    if (!(data->bound_dv = bound))
      data->use_get_bound = use_get_bound;

    data->info     = info;

    precon->precon_data = data;
    precon->init_precon = init_BPX_precon;
    precon->precon      = BPX_precon_s;
    precon->exit_precon = exit_BPX_precon;

    return(precon);
}

const PRECON *get_BPX_precon_d(const FE_SPACE *fe_space,
			       const DOF_SCHAR_VEC *bound,
			       int use_get_bound, int info)
{
    PRECON  *precon;
    HB_DATA *data;
    
    if (bound && (bound->fe_space != fe_space)) {
	ERROR("different fe spaces ?\n");
	return(nil);
    }

    precon = MEM_CALLOC(1, PRECON);
    data   = MEM_CALLOC(1, HB_DATA);
    
    data->fe_space = fe_space;
    if (!(data->bound_dv = bound))
      data->use_get_bound = use_get_bound;

    data->info     = info;

    precon->precon_data = data;
    precon->init_precon = init_BPX_precon;
    precon->precon      = BPX_precon_d;
    precon->exit_precon = exit_BPX_precon;

    return(precon);
}

/****************************************************************************/
