/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     Common/coarsen_common.c                                        */
/*                                                                          */
/*                                                                          */
/* description:  coarsening of 2 and 3 dim. hierarchical meshes;            */
/*               file is included by 2d/coarsen.c and 3d/coarsen.c;         */
/*               file contains all routines not depending on the dimension; */
/*                                                                          */
/* global variables that have to be defined in 2d/coarsen.c and 3d/coarsen.c*/
/*  static MESH           *coarse_mesh;                                     */
/*  static int            do_more;                                          */
/*  static int            call_coarse_restrict;                             */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* history:      first final version: 96/11/10                              */
/*                                    98/01/11  FE_SPACE structure          */
/*--------------------------------------------------------------------------*/

static int  count_coarse_restrict(MESH *mesh);

#if DIM == 1
U_CHAR coarsen(MESH *mesh)
{
  MACRO_EL *mel;
  int      n_elements = mesh->n_elements;

  call_coarse_restrict = count_coarse_restrict(mesh);

  for (mel = mesh->first_macro_el; mel; mel = mel->next)
    coarse_recursive(mesh, mel->el);

  return(mesh->n_elements > n_elements ? MESH_COARSENED : 0);
}
#else
/*--------------------------------------------------------------------------*/
/*  Propagate coarsening information over the whole hierarchy		    */
/*  by POSTORDER traversal of the hierarchy tree               		    */
/*  leaves:      'increment' coarsening mark                  		    */
/*  inner nodes: set coarsening mark to                                     */
/*               min(0,child[0].mark+1,child[1].mark+1)                     */
/*--------------------------------------------------------------------------*/

static void spread_coarsen_mark_fct(const EL_INFO *el_info)
{
  EL      *el = el_info->el;
  S_CHAR  mark;

  if (el->child[0])
  {  
/*--------------------------------------------------------------------------*/
/* interior node of the tree                                                */
/*--------------------------------------------------------------------------*/
    mark = MAX(el->child[0]->mark, el->child[1]->mark);
    el->mark = MIN(mark + 1, 0);
  }
  else
  {
/*--------------------------------------------------------------------------*/
/* leaf node of the tree                                                    */
/*--------------------------------------------------------------------------*/
    if (el->mark < 0)  el->mark -= 1;
  }
  return;
}

static void spread_coarsen_mark(MESH *mesh)
{
  mesh_traverse(mesh, -1, CALL_EVERY_EL_POSTORDER, spread_coarsen_mark_fct);
  return;
}

/*--------------------------------------------------------------------------*/
/*  cleanup_after_coarsen:                                   		    */
/*  resets the element marks                                                */
/*--------------------------------------------------------------------------*/

static void cleanup_after_coarsen_fct(const EL_INFO *el_info)
{
  EL       *el = el_info->el;
  el->mark = MAX(el->mark, 0);
}

static void cleanup_after_coarsen(MESH *mesh)
{
  mesh_traverse(mesh, -1, CALL_LEAF_EL, cleanup_after_coarsen_fct);
  return;
}

/*--------------------------------------------------------------------------*/
/*  do_coarse_patch:  if patch can be coarsend return true, else false and  */
/*  reset the element marks                                                 */
/*--------------------------------------------------------------------------*/

static int  do_coarse_patch(RC_LIST_EL *coarse_list, int n_neigh)
{
  FUNCNAME("do_coarse_patch");
  int      i, j;
  EL       *el;

  for (i = 0; i < n_neigh; i++)
  {
    el = coarse_list[i].el;

    if (el->mark >= 0  ||  el->child[0] == nil)
    {
/*--------------------------------------------------------------------------*/
/*  element must not be coarsend or element is a leaf element; reset the    */
/*  the coarsening flag on all those elements that have to be coarsend with */
/*  this element                                                            */
/*--------------------------------------------------------------------------*/
      el->mark = 0;
      for (j = 0; j < n_neigh; j++)
	if (coarse_list[j].flag)  coarse_list[j].el->mark = 0;
      return(false);
    }
    else if (el->child[0]->mark >= 0  ||  el->child[1]->mark >= 0)
    {
/*--------------------------------------------------------------------------*/
/*  one of the element's children must not be coarsend; reset the coarsening*/
/*  flag on all those elements that have to be coarsend with this element   */
/*--------------------------------------------------------------------------*/
      el->mark = 0;
      for (j = 0; j < n_neigh; j++)
	if (coarse_list[j].flag)  coarse_list[j].el->mark = 0;
      return(false);
    }
    else if (el->child[0]->child[0]  ||  el->child[1]->child[0])
    {
/*--------------------------------------------------------------------------*/
/*  one of the element's children is not a leaf element;                    */
/*  element may be coarsend after coarsening one of the children; try again */
/*--------------------------------------------------------------------------*/
      do_more = true;
      return(false);
    }
    else
    {
/*--------------------------------------------------------------------------*/
/*  either one element is a macro element or we can coarsen the patch       */
/*--------------------------------------------------------------------------*/
      if (coarse_list[i].flag == 0)
      {
	MACRO_EL  *mel;

	for (mel = coarse_mesh->first_macro_el; mel; mel = mel->next)
	  if (mel->el == el)  break;

	TEST_EXIT(mel)("incompatible coarsaning patch found\n");
      }
    }
  }

  return(true);
}


/*--------------------------------------------------------------------------*/
/*  coarsen_fct:  get the coarsen patch, starting on el_info->el, checks    */
/*  whether patch can be coarsend and calls coarse_patch if possible        */
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*  coarse_patch:  gets the coarsening patch via get_coarse_patch()         */
/*  (starting on el_info->el), checks wheter patch can be coarsend or not   */
/*  via do_coarsen_patch(); in the first case the patch is refined by       */
/*  coarse_patch(); otherwise element marker for patch elements are reset by*/
/*  do_coarsen_patch() and we have to return :-(                            */
/*--------------------------------------------------------------------------*/

static void coarsen_fct(const EL_INFO *el_info)
{
  EL          *el = el_info->el;
  DOF         *edge[2];
  int         n_neigh, bound = 0;
#if DIM == 2
  RC_LIST_EL  coarse_list[2];
#else
  EL_INFO     *elinfo = (EL_INFO *) el_info;
  RC_LIST_EL  *coarse_list;
  extern void get_neigh_on_patch(RC_LIST_EL [], int n_neigh, int bound);
#endif

  if (el->mark >= 0)  return;    /* el must not be coarsend, return :-(     */
  if (!(el->child[0])) return;   /* single leaves don't get coarsened       */

  if (el->child[0]->mark >= 0  || el->child[1]->mark >= 0)
  {
/*--------------------------------------------------------------------------*/
/*  one of the children must not be coarsend; return :-(                    */
/*--------------------------------------------------------------------------*/
    el->mark = 0;
    return;
  }

  if (el->child[0]->child[0] || el->child[1]->child[0])
  {
/*--------------------------------------------------------------------------*/
/*  one of the children is not a leaf element; try again later on           */
/*--------------------------------------------------------------------------*/
    do_more = true;
    return;
  }

#if DIM == 3
/*--------------------------------------------------------------------------*/
/*  get a list for storing all elements at the coarsening edge and fill it  */
/*--------------------------------------------------------------------------*/
  coarse_list = get_rc_list(coarse_mesh);
#endif

/*--------------------------------------------------------------------------*/
/*  give the refinement edge the right orientation                          */
/*--------------------------------------------------------------------------*/

  if (el->dof[0][0] < el->dof[1][0])
  {
    edge[0] = el->dof[0];
    edge[1] = el->dof[1];
  }
  else
  {
    edge[1] = el->dof[0];
    edge[0] = el->dof[1];
  }

  coarse_list[0].el = el;
  coarse_list[0].flag = 1;
  n_neigh = 1;
  
#if DIM == 2
  if ((coarse_list[1].el = NEIGH(el,el_info)[2]))
  {
    n_neigh = 2;
    coarse_list[1].flag = OPP_VERTEX(el,el_info)[2] == 2;
  }
#else
  coarse_list[0].opp_vertex[0] = 0;
  coarse_list[0].el_type = EL_TYPE(el,el_info);
  bound = false;
  if (get_coarse_patch(elinfo, edge, 0, coarse_list, &n_neigh))
  {
    get_coarse_patch(elinfo, edge, 1, coarse_list, &n_neigh);
    bound = true;
  }
  get_neigh_on_patch(coarse_list, n_neigh, bound);
#endif

/*--------------------------------------------------------------------------*/
/*  check wether we can coarsen the patch or not                            */
/*--------------------------------------------------------------------------*/
  if (do_coarse_patch(coarse_list, n_neigh))
  {
    coarse_patch(coarse_list, n_neigh, bound);
  }

#if DIM == 3
  free_rc_list(coarse_mesh, coarse_list);
#endif

  return;
}


/*--------------------------------------------------------------------------*/
/*  coarsen:         							    */
/*  traversal routine for recursiv coarsening of a triangulation            */
/*--------------------------------------------------------------------------*/

U_CHAR coarsen(MESH *mesh)
{
  int           n_elements;
#if NEIGH_IN_EL
  FLAGS         flag = CALL_EVERY_EL_POSTORDER;
#else
  FLAGS         flag = CALL_EVERY_EL_POSTORDER | FILL_NEIGH;
  const EL_INFO *el_info;
#endif

  n_elements = mesh->n_elements;
  coarse_mesh = mesh;

  call_coarse_restrict = count_coarse_restrict(mesh);

  spread_coarsen_mark(mesh);


#if NEIGH_IN_EL
  do
  {
    do_more = false;

    mesh_traverse(mesh, -1, flag, coarsen_fct);

  } while (do_more);
#else
  stack = get_traverse_stack();
  do
  {
    do_more = false;
    el_info = traverse_first(stack, mesh, -1, flag);
    while (el_info)
    {
      coarsen_fct(el_info);
      el_info = traverse_next(stack, el_info);
    }
  } while (do_more);
  free_traverse_stack(stack);
#endif
 
  cleanup_after_coarsen(mesh);

  n_elements -= mesh->n_elements;

  return(n_elements ? MESH_COARSENED : 0);
}
#endif  /*---  #if DIM == 1 */
/*--------------------------------------------------------------------------*/
/*  coarsening routines for global coarsening:				    */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  coarsen_mark_fct:							    */
/*  sets the mark on all elements that have to be coarsend                  */
/*--------------------------------------------------------------------------*/

static U_CHAR  global_mark;

static void coarsen_mark_fct(const EL_INFO *el_info)
{
#if 0
  el_info->el->mark = -(S_CHAR)MIN(global_mark,el_info->level);
#else
  el_info->el->mark = global_mark;
#endif
  return;
}

/*--------------------------------------------------------------------------*/
/*  global_coarsen:							    */
/*  tries to coarsen every element of mesh at least mark times		    */
/*--------------------------------------------------------------------------*/

U_CHAR global_coarsen(MESH *mesh, int mark)
{
  if (mark >= 0) return(0);
  global_mark = mark;
  mesh_traverse(mesh, -1, CALL_LEAF_EL, coarsen_mark_fct);
  return(coarsen(mesh));
}


/*--------------------------------------------------------------------------*/
/* restriction of dof vectors during mesh coarsening                        */
/* using DOF_VEC_LIST structure                                             */
/*--------------------------------------------------------------------------*/
typedef struct dof_vec_list     DOF_VEC_LIST;

struct dof_vec_list
{
  int             size;                       /* current size of total list */
  void            **list;          /* total list, distributed to ones below */

  int             n_dof_int_vec, n_dof_dof_vec, n_dof_uchar_vec,
                  n_dof_schar_vec, n_dof_real_vec, n_dof_real_d_vec,
                  n_dof_matrix;

  DOF_INT_VEC     **dof_int_vec;
  DOF_DOF_VEC     **dof_dof_vec;
  DOF_UCHAR_VEC   **dof_uchar_vec;
  DOF_SCHAR_VEC   **dof_schar_vec;
  DOF_REAL_VEC    **dof_real_vec;
  DOF_REAL_D_VEC  **dof_real_d_vec;
  DOF_MATRIX      **dof_matrix;
};

static DOF_VEC_LIST   dvlist[1] = {{0,nil}};

static int count_coarse_restrict(MESH *mesh)
{
  FUNCNAME("count_coarse_restrict");
  DOF_ADMIN      *admin;
  int            iadmin;
  DOF_REAL_VEC   *drv;
  DOF_REAL_D_VEC *drdv;
  DOF_INT_VEC    *div;
  DOF_DOF_VEC    *ddv;
  DOF_UCHAR_VEC  *duv;
  DOF_SCHAR_VEC  *dsv;
  DOF_MATRIX     *dm;
  int            ncr = 0;
  int            n_dof_int_vec=0,  n_dof_dof_vec=0,  n_dof_uchar_vec=0,
                 n_dof_schar_vec=0, n_dof_real_vec=0, n_dof_real_d_vec=0,
                 n_dof_matrix=0;

  for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
  {
    admin = mesh->dof_admin[iadmin];

    for (div = admin->dof_int_vec; div; div = div->next)
    {
      if (div->coarse_restrict) n_dof_int_vec++;
    }

    for (ddv = admin->dof_dof_vec; ddv; ddv = ddv->next)
    {
      if (ddv->coarse_restrict) n_dof_dof_vec++;
    }

    for (ddv = admin->int_dof_vec; ddv; ddv = ddv->next)
    {
      if (ddv->coarse_restrict) n_dof_int_vec++;
    }

    for (duv = admin->dof_uchar_vec; duv; duv = duv->next)
    {
      if (duv->coarse_restrict) n_dof_uchar_vec++;
    }

    for (dsv = admin->dof_schar_vec; dsv; dsv = dsv->next)
    {
      if (dsv->coarse_restrict) n_dof_schar_vec++;
    }

    for (drv = admin->dof_real_vec; drv; drv = drv->next)
    {
      if (drv->coarse_restrict) n_dof_real_vec++;
    }

    for (drdv = admin->dof_real_d_vec; drdv; drdv = drdv->next)
    {
      if (drdv->coarse_restrict) n_dof_real_d_vec++;
    }

    for (dm = admin->dof_matrix; dm; dm = dm->next)
    {
      if (dm->coarse_restrict) n_dof_matrix++;
    }
  }

  ncr = n_dof_int_vec + n_dof_dof_vec + n_dof_uchar_vec + n_dof_schar_vec
    + n_dof_real_vec + n_dof_real_d_vec + n_dof_matrix;

  if (ncr > 0)
  {
    if (dvlist->size < ncr) {
      dvlist->list = MEM_REALLOC(dvlist->list, dvlist->size, ncr+5, void *);
      dvlist->size = ncr+5;
    }

    ncr = 0;

    if (n_dof_int_vec)
      dvlist->dof_int_vec    = (DOF_INT_VEC **)(dvlist->list+ncr);
    else
      dvlist->dof_int_vec    = nil;
    ncr += n_dof_int_vec;

    if (n_dof_dof_vec)
      dvlist->dof_dof_vec    = (DOF_DOF_VEC **)(dvlist->list+ncr);
    else
      dvlist->dof_dof_vec    = nil;
    ncr += n_dof_dof_vec;

    if (n_dof_uchar_vec)
      dvlist->dof_uchar_vec  = (DOF_UCHAR_VEC **)(dvlist->list+ncr);
    else
      dvlist->dof_uchar_vec  = nil;
    ncr += n_dof_uchar_vec;

    if (n_dof_schar_vec)
      dvlist->dof_schar_vec  = (DOF_SCHAR_VEC **)(dvlist->list+ncr);
    else
      dvlist->dof_schar_vec  = nil;
    ncr += n_dof_schar_vec;

    if (n_dof_real_vec)
      dvlist->dof_real_vec   = (DOF_REAL_VEC **)(dvlist->list+ncr);
    else
      dvlist->dof_real_vec   = nil;
    ncr += n_dof_real_vec;

    if (n_dof_real_d_vec)
      dvlist->dof_real_d_vec = (DOF_REAL_D_VEC **)(dvlist->list+ncr);
    else
      dvlist->dof_real_d_vec = nil;
    ncr += n_dof_real_d_vec;

    if (n_dof_matrix)
      dvlist->dof_matrix     = (DOF_MATRIX **)(dvlist->list+ncr);
    else
      dvlist->dof_matrix     = nil;
    ncr += n_dof_matrix;
    TEST_EXIT(ncr <= dvlist->size)("error in dvlist->size");

    dvlist->n_dof_int_vec    = 0;
    dvlist->n_dof_dof_vec    = 0;
    dvlist->n_dof_uchar_vec  = 0;
    dvlist->n_dof_schar_vec  = 0;
    dvlist->n_dof_real_vec   = 0;
    dvlist->n_dof_real_d_vec = 0;
    dvlist->n_dof_matrix     = 0;

    for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
    {
      admin = mesh->dof_admin[iadmin];
      for (div = admin->dof_int_vec; div; div = div->next) {
	if (div->coarse_restrict)
	  dvlist->dof_int_vec[dvlist->n_dof_int_vec++] = div;
      }
      for (ddv = admin->dof_dof_vec; ddv; ddv = ddv->next) {
	if (ddv->coarse_restrict) 
	  dvlist->dof_dof_vec[dvlist->n_dof_dof_vec++] = ddv;
      }
      for (ddv = admin->int_dof_vec; ddv; ddv = ddv->next) {
	if (ddv->coarse_restrict) 
	  dvlist->dof_dof_vec[dvlist->n_dof_dof_vec++] = ddv;
      }
      for (duv = admin->dof_uchar_vec; duv; duv = duv->next) {
	if (duv->coarse_restrict)
	  dvlist->dof_uchar_vec[dvlist->n_dof_uchar_vec++] = duv;
      }
      for (dsv = admin->dof_schar_vec; dsv; dsv = dsv->next) {
	if (dsv->coarse_restrict) 
	  dvlist->dof_schar_vec[dvlist->n_dof_schar_vec++] = dsv;
      }
      for (drv = admin->dof_real_vec; drv; drv = drv->next) {
	if (drv->coarse_restrict)
	  dvlist->dof_real_vec[dvlist->n_dof_real_vec++] = drv;
      }
      for (drdv = admin->dof_real_d_vec; drdv; drdv = drdv->next) {
	if (drdv->coarse_restrict) 
	  dvlist->dof_real_d_vec[dvlist->n_dof_real_d_vec++] = drdv;
      }
      for (dm = admin->dof_matrix; dm; dm = dm->next) {
	if (dm->coarse_restrict) 
	  dvlist->dof_matrix[dvlist->n_dof_matrix++] = dm;
      }
    }

    TEST_EXIT(dvlist->n_dof_int_vec == n_dof_int_vec)
      ("error in n_dof_int_vec");
    TEST_EXIT(dvlist->n_dof_dof_vec == n_dof_dof_vec)
      ("error in n_dof_dof_vec");
    TEST_EXIT(dvlist->n_dof_uchar_vec == n_dof_uchar_vec)
      ("error in n_dof_uchar_vec");
    TEST_EXIT(dvlist->n_dof_schar_vec == n_dof_schar_vec)
      ("error in n_dof_schar_vec");
    TEST_EXIT(dvlist->n_dof_real_vec == n_dof_real_vec)
      ("error in n_dof_real_vec");
    TEST_EXIT(dvlist->n_dof_real_d_vec == n_dof_real_d_vec)
      ("error in n_dof_real_d_vec");
    TEST_EXIT(dvlist->n_dof_matrix == n_dof_matrix)
      ("error in n_dof_matrix");
  }
  else
  {
    dvlist->dof_int_vec    = nil;
    dvlist->dof_dof_vec    = nil;
    dvlist->dof_uchar_vec  = nil;
    dvlist->dof_schar_vec  = nil;
    dvlist->dof_real_vec   = nil;
    dvlist->dof_real_d_vec = nil;
    dvlist->dof_matrix     = nil;
  }

  return(ncr);
}

static void coarse_restrict(RC_LIST_EL *list, int n_el)
{
  DOF_REAL_VEC    *drv;
  DOF_REAL_D_VEC  *drdv;
  DOF_INT_VEC     *div;
  DOF_DOF_VEC     *ddv;
  DOF_UCHAR_VEC   *duv;
  DOF_SCHAR_VEC   *dsv;
  DOF_MATRIX      *dm;
  int             i;

  if (!dvlist) return;

  for (i = 0; i < dvlist->n_dof_int_vec; i++) 
  {
    div = dvlist->dof_int_vec[i];
    if (div->coarse_restrict)  div->coarse_restrict(div, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_dof_vec; i++) 
  {
    ddv = dvlist->dof_dof_vec[i];
    if (ddv->coarse_restrict)  ddv->coarse_restrict(ddv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_uchar_vec; i++) 
  {
    duv = dvlist->dof_uchar_vec[i];
    if (duv->coarse_restrict)  duv->coarse_restrict(duv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_schar_vec; i++) 
  {
    dsv = dvlist->dof_schar_vec[i];
    if (dsv->coarse_restrict)  dsv->coarse_restrict(dsv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_real_vec; i++) 
  {
    drv = dvlist->dof_real_vec[i];
    if (drv->coarse_restrict)  drv->coarse_restrict(drv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_real_d_vec; i++) 
  {
    drdv = dvlist->dof_real_d_vec[i];
    if (drdv->coarse_restrict) drdv->coarse_restrict(drdv, list, n_el);
  }

  for (i = 0; i < dvlist->n_dof_matrix; i++) 
  {
    dm = dvlist->dof_matrix[i];
    if (dm->coarse_restrict) dm->coarse_restrict(dm, list, n_el);
  }

  return;
}

/*--------------------------------------------------------------------------*/
/*  calculate the maximal level of the triangulation                        */
/*--------------------------------------------------------------------------*/

static  unsigned int max_level_val;

static void max_level_fct(const EL_INFO *el_info)
{
  max_level_val = MAX(max_level_val, el_info->level);
}

int get_max_level(MESH *mesh)
{
  max_level_val = 0;
  mesh_traverse(mesh, -1, CALL_LEAF_EL, max_level_fct);
  return(max_level_val);
}
