/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     trav_xy.c                                                      */
/*                                                                          */
/*                                                                          */
/* description: Find element at position (xy) and local coordinates         */
/*              return 0 if xy \not\in domain.                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Univesitaet Bremen                                           */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*             Oliver Kriessl                                               */
/*             Institut fuer Mathematik                                     */
/*             Albert-Ludwigs-Universitaet Freiburg                         */
/*             Hermann-Herder-Str. 10                                       */
/*             D-79104 Freiburg                                             */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

static const REAL *g_xy0, *g_xy;
static REAL       *g_sp;
static EL_INFO final_el_info[1];
static REAL   final_lambda[DIM+1] = {0.0};

static int find_el_at_pt_recursive(EL_INFO *el_info, REAL lambda[DIM+1],
				   int outside)
{
  FUNCNAME("find_el_at_pt_recursive");
  EL      *el = el_info->el;
  EL_INFO c_el_info[1];
  REAL    c_lambda[DIM+1];
  int     i, inside;
#if DIM == 3
  int     ichild, c_outside;
#endif /* DIM == 3 */

#if (DIM == 3) && 0
  if (outside >= 0) {
    MSG("el %2d: leaf=%d, outside=%d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	INDEX(el), IS_LEAF_EL(el), outside,
	lambda[0],lambda[1],lambda[2],lambda[3]);
  }
  else {
    MSG("el %2d: leaf=%d, inside, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	INDEX(el), IS_LEAF_EL(el), lambda[0],lambda[1],lambda[2],lambda[3]);
  }
#endif /* (DIM == 3) && 0 */

  if (IS_LEAF_EL(el)) {
    *final_el_info = *el_info;
    if (outside < 0) {
      for (i=0; i<=DIM; i++)  final_lambda[i] = lambda[i];

/* ======================= first 0 conditional ============================ */
#if 0 
      for (i=0; i<=DIM; i++)  
      {
	if (final_lambda[i] < -1.0E-5)
#if DIM == 1
	  ERROR("outside element: %.5lf %.5lf\n",
		lambda[0],lambda[1]);
#endif /* DIM == 1 */
#if DIM == 2
	  ERROR("outside element: %.5lf %.5lf %.5lf\n",
		lambda[0],lambda[1],lambda[2]);
#endif /* DIM == 2 */
#if DIM == 3
	  ERROR("outside element: %.5lf %.5lf %.5lf %.5lf\n",
		lambda[0],lambda[1],lambda[2],lambda[3]);
#endif /* DIM == 3 */
      }
#endif 
/* ===================== end of first 0 conditional ======================= */

/* ======================= second 0 conditional =========================== */
#if 0
      {
	REAL_D world;
	REAL   s;

	coord_to_world(el_info, lambda, world);
	s = 0.0;
	for (i=0; i<DIM; i++) s += SQR(world[i]-g_xy[i]);

	if (s > 1.E-10) 
	{
	  MSG("el %d:\n", INDEX(el));
#if DIM == 1
	  MSG("? xy=(%.2le), world=(%.2le)\n", g_xy[0], world[0]);
#endif /* DIM == 1 */
#if DIM == 2
	  MSG("? xy=(%.2le, %.2le), world=(%.2le, %.2le)\n",
	      g_xy[0], g_xy[1], world[0], world[1]);
#endif /* DIM == 2 */
#if DIM == 3
	  MSG("? xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	      g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
#endif /* DIM == 3 */
	  WAIT_REALLY;
	}
      }
#endif
/* ===================== end of second 0 conditional ====================== */

      return(true);
    }
    else 
    {  /* outside */
      if (g_xy0) 
      { /* find boundary point of [xy0, xy] */
	REAL s;
	world_to_coord(el_info, g_xy0, c_lambda);
	s = lambda[outside] / (lambda[outside] - c_lambda[outside]);
	for (i=0; i<=DIM; i++) 
	{
	  final_lambda[i] = s * c_lambda[i] + (1.0-s) * lambda[i];
	}
	if (g_sp) *g_sp = s;
#if DIM == 3
	MSG("outside finest level on el %d: s=%.3le\n", INDEX(el), s);
#endif /* DIM == 3 */
	return(false);  /* ??? */
      }
      else return(false);
    }
  }

#if DIM == 1
  if (lambda[0] >= lambda[1]) {
    fill_elinfo(0, el_info, c_el_info);
    if (outside >= 0) 
    {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      if (outside >= 0) ERROR("point outside domain\n");
    }
    else {
      c_lambda[0] = lambda[0] - lambda[1];
      c_lambda[1] = 2.0 * lambda[1];
    }
  }
  else {
    fill_elinfo(1, el_info, c_el_info);
    if (outside >= 0) 
    {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      if (outside >= 0) ERROR("point outside domain\n");
    }
    else {
      c_lambda[1] = lambda[1] - lambda[0];
      c_lambda[0] = 2.0 * lambda[0];
    }
  }
#endif /* DIM == 1 */
#if DIM == 2
  if (lambda[0] >= lambda[1]) {
    fill_elinfo(0, el_info, c_el_info);
    if (el->new_coord) {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      if (outside >= 0) {
	ERROR("outside curved boundary child 0\n");
      }
    }
    else {
      c_lambda[0] = lambda[2];
      c_lambda[1] = lambda[0] - lambda[1];
      c_lambda[2] = 2.0 * lambda[1];
    }
  }
  else {
    fill_elinfo(1, el_info, c_el_info);
    if (el->new_coord) {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      if (outside >= 0) {
	ERROR("outside curved boundary child 1\n");
      }
    }
    else {
      c_lambda[0] = lambda[1] - lambda[0];
      c_lambda[1] = lambda[2];
      c_lambda[2] = 2.0 * lambda[0];
    }
  }

#endif /* DIM == 2 */
#if DIM == 3 

  if (el->new_coord) {
    if (lambda[0] >= lambda[1])
      ichild = 0;
    else
      ichild = 1;
    fill_elinfo(ichild, el_info, c_el_info);
    c_outside = world_to_coord(c_el_info, g_xy, c_lambda);

    if (c_outside>=0) {  /* test is other child is better... */
      REAL c_lambda2[DIM+1];
      int  c_outside2;
      EL_INFO c_el_info2[1];

      fill_elinfo(1-ichild, el_info, c_el_info2);
      c_outside2 = world_to_coord(c_el_info2, g_xy, c_lambda2);

      MSG("new_coord CHILD %d: outside=%d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	  ichild, c_outside, c_lambda[0],c_lambda[1],c_lambda[2],c_lambda[3]);
      MSG("new_coord CHILD %d: outside=%d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
    1-ichild, c_outside2, c_lambda2[0],c_lambda2[1],c_lambda2[2],c_lambda2[3]);

      /* if (c_lambda[0] < c_lambda2[0]) { */
      if ((c_outside2 < 0) || (c_lambda2[c_outside2] > c_lambda[c_outside])) {
	for (i=0; i<=DIM; i++) c_lambda[i] = c_lambda2[i];
	c_outside = c_outside2;
	*c_el_info = *c_el_info2;
	ichild = 1 - ichild;
      }
    }
    outside = c_outside;

/* ==================== first 0 conditional in 3d-part ==================== */
#if 0
    MSG("new_coord child %d: el %d: outside=%2d, ", ichild, 
	INDEX(el->child[0]), outside);
    MSG("lambda=(%.2lf %.2lf %.2lf %.2lf)\n", c_lambda[0],c_lambda[1],
	c_lambda[2],c_lambda[3]);
    {
      REAL_D world;
      coord_to_world(c_el_info, c_lambda, world);
      MSG("xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	  g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
    }
#endif
/* ================ end of first 0 conditional in 3d-part ================= */

    /* if (outside >= 0) ERROR("outside curved boundary child 0\n"); */
  }
  else {  /* no new_coord */
    if (lambda[0] >= lambda[1]) {
      fill_elinfo(0, el_info, c_el_info);
      MSG("transform lambda child 0\n");
      c_lambda[0] = lambda[0] - lambda[1];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][0][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][0][2]];
      c_lambda[3] = 2.0 * lambda[1];
    }
    else {
      fill_elinfo(1, el_info, c_el_info);
      MSG("transform lambda child 1\n");
      c_lambda[0] = lambda[1] - lambda[0];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][1][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][1][2]];
      c_lambda[3] = 2.0 * lambda[0];
    }
  }

/* =================== second 0 conditional in 3d-part ==================== */
#if 0
  if (lambda[0] >= lambda[1]) 
  {
    fill_elinfo(0, el_info, c_el_info);
    if (el->new_coord) 
    {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      MSG("new_coord child 0: el %d: outside=%2d, ", 
	  INDEX(el->child[0]), outside);
      MSG("lambda=(%.2lf %.2lf %.2lf %.2lf)\n", c_lambda[0], 
	  c_lambda[1], c_lambda[2], c_lambda[3]);
      /* if (outside >= 0) ERROR("outside curved boundary child 0\n"); */
    }
    else 
    {
      MSG("transform lambda child 0\n");
      c_lambda[0] = lambda[0] - lambda[1];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][0][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][0][2]];
      c_lambda[3] = 2.0 * lambda[1];
    }
  } /* end "if (lambda[0] >= lambda[1])" */
  else 
  {
    fill_elinfo(1, el_info, c_el_info);
    if (el->new_coord) 
    {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      MSG("new_coord child 1: el %d: outside=%2d, ",
	  INDEX(el->child[1]), outside);
      MSG("lambda=(%.2lf %.2lf %.2lf %.2lf)\n", c_lambda[0],
	  c_lambda[1], c_lambda[2], c_lambda[3]);
      {
	REAL_D world;
	coord_to_world(c_el_info, c_lambda, world);
	MSG("xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	    g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
      }
      /* if (outside >= 0) ERROR("outside curved boundary child 1\n"); */
    }
    else 
    {
      MSG("transform lambda child 1\n");
      c_lambda[0] = lambda[1] - lambda[0];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][1][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][1][2]];
      c_lambda[3] = 2.0 * lambda[0];
    }
  } /* end of else of "if (lambda[0] >= lambda[1])" */

  {
    REAL_D world;
    coord_to_world(c_el_info, c_lambda, world);
    MSG("xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
  }
#endif
/* =============== end of second 0 conditional in 3d-part ================= */

#endif  /* DIM == 3 */

  inside = find_el_at_pt_recursive(c_el_info, c_lambda, outside);
  return(inside);
}

/*--------------------------------------------------------------------------*/

int find_el_at_pt(MESH *mesh, const REAL_D xy, EL_INFO **el_info_p, FLAGS flag, 
		  REAL bary[DIM+1], const MACRO_EL *start_mel, const REAL_D xy0,
		  REAL *sp)
{
  const MACRO_EL  *mel;
  REAL      lambda[DIM+1];
  EL_INFO   mel_info[1];
  int       i, k, inside;

  TEST_EXIT(el_info_p)("need pointer to pointer to an el_info structure\n");

  if (start_mel)
    mel = start_mel;
  else
    mel = mesh->first_macro_el;

  mel_info->fill_flag = flag|FILL_COORDS;
  g_xy  = xy;
  g_xy0 = xy0;
  g_sp  = sp;

  fill_macro_info(mesh, mel, mel_info);

  while ((k = world_to_coord(mel_info, xy, lambda)) >= 0) {
    if (mel->neigh[k]) {
      mel = mel->neigh[k];
      fill_macro_info(mesh, mel, mel_info);
      continue;
    }

#if (DIM > 1) && 0
    if (xy0) 
    { /* find boundary point of [xy0, xy] */
      static REAL s;

      if (mel_info->el->new_coord) {
	MSG("may produce incorrect results near curved boundary\n");
      }

      world_to_coord(mel_info, xy0, bary);
      s = lambda[k] / (lambda[k] - bary[k]);
      for (i=0; i<=DIM; i++) {
	lambda[i] = s * bary[i] + (1.0-s) * lambda[i];
      }
      if (sp) *sp = s;
    }
    else {
      return(false);
    }
#endif

    break;
  }

#if 0
  if (k >= 0) {
    MSG("outside=%d on macro level.\n", k);
  }
  else {
    MSG("inside on macro level.\n", k);
  }
#endif

  /* now, descend in tree to find leaf element at point */
  inside = find_el_at_pt_recursive(mel_info, lambda, k);
  for (i=0; i<=DIM; i++) bary[i] = final_lambda[i];
  *el_info_p = final_el_info;
  
  /* if ((k>=0) || (!inside)) WAIT; */

  return(inside);
}
