/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     write_mesh.c                                                   */
/*                                                                          */
/*                                                                          */
/* description:  functions for writing meshes and DOF vectors in binary     */
/*               machine-dependent format                                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Univesitaet Bremen                                           */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* to_do :  number of dofs depending on node                                */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

static FILE *file = nil;
static int  *n_dof, *node;
static U_CHAR preserve_coarse_dofs;

static void write_el_recursive(EL *el);

static int         n_vert_ptrs;
static DOF         *vert_dofs, **vert_ptrs;
static DOF_ADMIN   *vert_admin;

#if DIM > 1
static int         n_edge_ptrs;
static DOF         *edge_dofs, **edge_ptrs;
static DOF_ADMIN   *edge_admin;
#endif

#if DIM==3
static int         n_face_ptrs;
static DOF         *face_dofs, **face_ptrs;
static DOF_ADMIN   *face_admin;
#endif

static int         *mesh_n_dof, *mesh_node;

/*--------------------------------------------------------------------------*/
#define OFFSET 10      /*---  must be > MAX(VERTEX,EDGE,FACE,CENTER)     ---*/
/*--------------------------------------------------------------------------*/

void fill_dofs_fct(const EL_INFO *elinfo)
{
  EL   *el = elinfo->el;
  int  i, n0, n1;
  DOF  *p;
  
  if (IS_LEAF_EL(el) || preserve_coarse_dofs)
  {
    if (vert_admin) {
      n0 = mesh_node[VERTEX];
      n1 = vert_admin->n0_dof[VERTEX];
      for (i = 0; i < N_VERTICES; i++) {
	p = el->dof[n0+i];
	vert_dofs[p[n1]] = VERTEX-OFFSET;
	vert_ptrs[p[n1]] = p;
      }
    }
#if DIM > 1
    if (edge_admin) {
      n0 = mesh_node[EDGE];
      n1 = edge_admin->n0_dof[EDGE];
      for (i = 0; i < N_EDGES; i++) {
	p = el->dof[n0+i];
	edge_dofs[p[n1]] = EDGE-OFFSET;
	edge_ptrs[p[n1]] = p;
      }
    }
#endif

#if DIM == 3
    if (face_admin) {
      n0 = mesh_node[FACE];
      n1 = face_admin->n0_dof[FACE];
      for (i = 0; i < N_FACES; i++) {
	p = el->dof[n0+i];
	face_dofs[p[n1]] = FACE-OFFSET;
	face_ptrs[p[n1]] = p;
      }
    }
#endif
  }
  return;
}
/*--------------------------------------------------------------------------*/

int write_mesh(MESH *mesh, const char *filename, REAL time)
{
  FUNCNAME("write_mesh");
  MACRO_EL   *mel, *meln;
  DOF_ADMIN  *admin;
  int        i, j, n, m, iadmin;
  int        neigh_i[N_NEIGH];
#if DIM > 1
  S_CHAR     bound_sc[N_NEIGH];
#endif
#if DIM==3
  S_CHAR     edge_sc[N_EDGES];
#endif

  if (!mesh)
  {
    ERROR("no mesh - no file created\n");
    return(1);
  }

  dof_compress(mesh);

  n_dof = mesh->n_dof;
  node  = mesh->node;

  if (!(file=fopen(filename,"wb")))
  {
    ERROR("cannot open file %s\n",filename);
    return(1);
  }

  fwrite(ALBERTA_VERSION, sizeof(char), strlen(ALBERTA_VERSION)+1, file);  /* file marker */

  i = sizeof(REAL);
  fwrite(&i, sizeof(int), 1, file);

  i = DIM;
  fwrite(&i, sizeof(int), 1, file);

  i = DIM_OF_WORLD;
  fwrite(&i, sizeof(int), 1, file);

  fwrite(&time, sizeof(REAL), 1, file);

  i = mesh->name ? strlen(mesh->name) : 0;  /* length without terminating \0 */
  fwrite(&i, sizeof(int), 1, file);
  fwrite(mesh->name, sizeof(char), i, file);

  fwrite(mesh->diam, sizeof(REAL), DIM_OF_WORLD, file);
  fwrite(&mesh->preserve_coarse_dofs, sizeof(U_CHAR), 1, file);
  preserve_coarse_dofs = mesh->preserve_coarse_dofs;
  
  fwrite(&(mesh->n_dof_el), sizeof(int), 1, file);
  fwrite(mesh->n_dof, sizeof(int), DIM+1, file);
  fwrite(&(mesh->n_node_el), sizeof(int), 1, file);
  fwrite(mesh->node, sizeof(int), DIM+1, file);

  fwrite(&(mesh->n_dof_admin), sizeof(int), 1, file);
  for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
  {
    admin = mesh->dof_admin[iadmin];
    if (admin)
    {
      fwrite(admin->n_dof, sizeof(int), DIM+1, file);
      fwrite(&(admin->used_count), sizeof(int), 1, file);
         /* after dof_compress(), no more information is required */
      i = admin->name ? strlen(admin->name) : 0;
         /* length of name without terminating \0 */
      fwrite(&i, sizeof(int), 1, file);
      fwrite(admin->name, sizeof(char), i, file);
    }
    else
    {
      i = 0;
      for (j = 0; j < DIM+3; j++) fwrite(&i, sizeof(int), 1, file);
    }
  }

#if DIM==3
  face_admin = nil;
  if (mesh->n_dof[FACE] > 0) {
    m = 0;
    for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++) {
      if ((admin = mesh->dof_admin[iadmin])) {
	if (admin->n_dof[FACE] > 0  && (m==0 || admin->n_dof[FACE] < m)) {
	  face_admin = admin;
	  m = admin->n_dof[FACE];
	}
      }
    }
    TEST_EXIT(face_admin)("no admin with face dofs?\n");
    face_dofs = MEM_ALLOC(face_admin->used_count, DOF);
    for (i=0; i<face_admin->used_count; i++) face_dofs[i] = -1;
    face_ptrs = MEM_ALLOC(face_admin->used_count, DOF*);
    for (i=0; i<face_admin->used_count; i++) face_ptrs[i] = 0;
  }
#endif  

#if DIM > 1
  edge_admin = nil;
  if (mesh->n_dof[EDGE] > 0)
  {
#if DIM==3
    if (face_admin && (face_admin->n_dof[EDGE] > 0)) {
      edge_admin = face_admin;
      edge_dofs = face_dofs;
      edge_ptrs = face_ptrs;
    }
    else
#endif
    {
      m = 0;
      for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++) {
	if ((admin = mesh->dof_admin[iadmin])) {
	  if (admin->n_dof[EDGE] > 0  && (m==0 || admin->n_dof[EDGE] < m)) {
	    edge_admin = admin;
	    m = admin->n_dof[EDGE];
	  }
	}
      }
      TEST_EXIT(edge_admin)("no admin with edge dofs?\n");
      edge_dofs = MEM_ALLOC(edge_admin->used_count, DOF);
      for (i=0; i<edge_admin->used_count; i++) edge_dofs[i] = -1;
      edge_ptrs = MEM_ALLOC(edge_admin->used_count, DOF*);
      for (i=0; i<edge_admin->used_count; i++) edge_ptrs[i] = 0;
    }
  }
#endif

  vert_admin = nil;
  if (mesh->n_dof[VERTEX] > 0)
  {
#if DIM == 3
    if (face_admin && (face_admin->n_dof[VERTEX] > 0)) 
    {
      vert_admin = face_admin;
      vert_dofs = face_dofs;
      vert_ptrs = face_ptrs;
    }
    else
#endif
#if DIM > 1
    if (edge_admin && (edge_admin->n_dof[VERTEX] > 0)) 
    {
      vert_admin = edge_admin;
      vert_dofs = edge_dofs;
      vert_ptrs = edge_ptrs;
    }
    else
#endif
    {
      m = 0;
      for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++) {
	if ((admin = mesh->dof_admin[iadmin])) {
	  if (admin->n_dof[VERTEX] > 0  && (m==0 || admin->n_dof[VERTEX] < m)) {
	    vert_admin = admin;
	    m = admin->n_dof[VERTEX];
	  }
	}
      }
      TEST_EXIT(vert_admin)("no admin with vertex dofs?\n");
      vert_dofs = MEM_ALLOC(vert_admin->used_count, DOF);
      for (i=0; i<vert_admin->used_count; i++) vert_dofs[i] = -1;
      vert_ptrs = MEM_ALLOC(vert_admin->used_count, DOF*);
      for (i=0; i<vert_admin->used_count; i++) vert_ptrs[i] = 0;
    }
  }

  mesh_n_dof = mesh->n_dof;
  mesh_node  = mesh->node;
  mesh_traverse(mesh, 0, CALL_EVERY_EL_PREORDER|FILL_NOTHING, fill_dofs_fct);

  if (vert_admin) 
  {
    n_vert_ptrs = 0;
    for (i = 0; i < vert_admin->used_count; i++)
      if (vert_dofs[i] == VERTEX-OFFSET)
	n_vert_ptrs++;
  }

#if DIM > 1
  if (edge_admin) 
  {
    n_edge_ptrs = 0;
    for (i = 0; i < edge_admin->used_count; i++)
      if (edge_dofs[i] == EDGE-OFFSET)
	n_edge_ptrs++;
  }
#endif

#if DIM==3
  if (face_admin)
  {
    n_face_ptrs = 0;
    for (i = 0; i < face_admin->used_count; i++)
      if (face_dofs[i] == FACE-OFFSET)
	n_face_ptrs++;
  }
#endif

  fwrite(&n_vert_ptrs, sizeof(int), 1, file);
  if (n_vert_ptrs) {
    j = 0;
    n = mesh_n_dof[VERTEX];
    for (i = 0; i < vert_admin->used_count; i++) {
      if (vert_dofs[i] == VERTEX-OFFSET) {
	vert_dofs[i] = j++;
	fwrite(vert_ptrs[i], sizeof(DOF), n, file);
      }
    }
    TEST_EXIT(j==n_vert_ptrs)("j != n_vert_ptrs\n");
  }
#if DIM > 1
  fwrite(&n_edge_ptrs, sizeof(int), 1, file);
  if (n_edge_ptrs)
  {
    j = 0;
    n = mesh_n_dof[EDGE];
    for (i = 0; i < edge_admin->used_count; i++) {
      if (edge_dofs[i] == EDGE-OFFSET) {
	edge_dofs[i] = j++;
	fwrite(edge_ptrs[i], sizeof(DOF), n, file);
      }
    }
    TEST_EXIT(j==n_edge_ptrs)("j != n_edge_ptrs\n");
  }
#endif

#if DIM==3
  fwrite(&n_face_ptrs, sizeof(int), 1, file);
  if (n_face_ptrs) {
    j = 0;
    n = mesh_n_dof[FACE];
    for (i = 0; i < face_admin->used_count; i++) {
      if (face_dofs[i] == FACE-OFFSET) {
	face_dofs[i] = j++;
	fwrite(face_ptrs[i], sizeof(DOF), n, file);
      }
    }
    TEST_EXIT(j==n_face_ptrs)("j != n_face_ptrs\n");
  }
#endif

/*--------------------------------------------------------------------------*/
/* gather info about macro elements (vertices, ...)                         */
/*--------------------------------------------------------------------------*/
  {
    typedef int  intNV[N_VERTICES];
    int    (*mcindex)[N_VERTICES] = MEM_ALLOC(mesh->n_macro_el, intNV);
    REAL_D *mccoord;
    int    mccount, m;

    mccount  = ((MESH_MEM_INFO *)mesh->mem_info)->count;
    mccoord  = ((MESH_MEM_INFO *)mesh->mem_info)->coords;

    for (m = 0, mel = mesh->first_macro_el;  mel;  m++, mel = mel->next)
    {
      for (i = 0; i < N_VERTICES; i++)
	mcindex[m][i] = 
((char *)mel->coord[i] - (char *)mccoord)/(sizeof(REAL_D)*sizeof(char));

      if (mel->index != m) mel->index = m;
    }
    TEST(m == mesh->n_macro_el)(" mesh->n_macro_el = %d  !=  %d ???\n");

    fwrite(&(mesh->n_macro_el), sizeof(int), 1, file);

    fwrite(&mccount, sizeof(int), 1, file);     /* number of macro coords */

    for (i = 0; i < mccount; i++)
      fwrite(mccoord[i], sizeof(REAL_D), 1, file);

    for (m = 0, mel = mesh->first_macro_el;  mel;  m++, mel = mel->next)
    {
      fwrite(mcindex[m], sizeof(int), N_VERTICES, file);

      fwrite(mel->bound, sizeof(S_CHAR), N_VERTICES, file);

#if DIM > 1
      for (i = 0; i < N_NEIGH; i++)
	bound_sc[i] = GET_BOUND(mel->boundary[i]);
      fwrite(bound_sc, sizeof(S_CHAR), N_NEIGH, file);
#endif

      for (i = 0; i < N_NEIGH; i++)
      {
	if ((meln = mel->neigh[i]))
	  neigh_i[i] = meln->index;
	else
	  neigh_i[i] = -1;
      }
      fwrite(neigh_i, sizeof(int), N_NEIGH, file);

      fwrite(mel->opp_vertex, sizeof(U_CHAR), N_NEIGH, file);
    
#if DIM == 3
      for (i = 0; i < N_EDGES; i++)
	edge_sc[i] = GET_BOUND(mel->boundary[N_FACES+i]);
      fwrite(edge_sc, sizeof(S_CHAR), N_EDGES, file); 
    /* zusammen mit FACE-BOUNDS ???  */

      fwrite(&(mel->el_type), sizeof(U_CHAR), 1, file);
#endif

      write_el_recursive(mel->el);
    }
    TEST_EXIT(m == mesh->n_macro_el)
      ("m %d != n_macro_el %d\n", m, mesh->n_macro_el);

    MEM_FREE(mcindex, mesh->n_macro_el, intNV);
  }

  fwrite(&mesh->n_vertices, sizeof(int), 1, file);
#if DIM > 1
  fwrite(&mesh->n_edges, sizeof(int), 1, file);
#endif
  fwrite(&mesh->n_elements, sizeof(int), 1, file);
  fwrite(&mesh->n_hier_elements, sizeof(int), 1, file);
#if DIM == 3
  fwrite(&mesh->n_faces, sizeof(int), 1, file);
  fwrite(&mesh->max_edge_neigh, sizeof(int), 1, file);
#endif
  fwrite("EOF.", 4, 1, file);                            /* file end marker */
  fclose(file);

  file = nil;

#if DIM==3
  if (face_admin) 
  {
    MEM_FREE(face_ptrs, face_admin->used_count, DOF*);
    MEM_FREE(face_dofs, face_admin->used_count, DOF);
    if (edge_admin == face_admin)  edge_admin = nil;
    if (vert_admin == face_admin)  vert_admin = nil;
    face_admin = nil;
  }
#endif
#if DIM > 1
  if (edge_admin) 
  {
    MEM_FREE(edge_ptrs, edge_admin->used_count, DOF*);
    MEM_FREE(edge_dofs, edge_admin->used_count, DOF);
    if (vert_admin == edge_admin)  vert_admin = nil;
    edge_admin = nil;
  }
#endif
  if (vert_admin) 
  {
    MEM_FREE(vert_ptrs, vert_admin->used_count, DOF*);
    MEM_FREE(vert_dofs, vert_admin->used_count, DOF);
    vert_admin = nil;
  }

  return(0);
}

/*--------------------------------------------------------------------------*/

static void write_el_recursive(EL *el)
{
  FUNCNAME("write_el_recursive");
  int    i, n, node0, n1;
  static U_CHAR uc_no = 0, uc_yes = 1;

  TEST_EXIT(el)("el == nil\n");

  if (el->child[0])
  {
    TEST_EXIT(el->child[1])("child 0 but no child 1\n");
    fwrite(&uc_yes, sizeof(U_CHAR), 1, file);
  }
  else
  {
    fwrite(&uc_no, sizeof(U_CHAR), 1, file);
  }

#if DIM > 1
  if (el->new_coord) 
  {
    fwrite(&uc_yes, sizeof(U_CHAR), 1, file);
    fwrite(el->new_coord, sizeof(REAL_D), 1, file);
  }
  else 
  {
    fwrite(&uc_no, sizeof(U_CHAR), 1, file);
  }
#endif

  if (n_dof[VERTEX] > 0)
  {
    node0 = node[VERTEX];
    n1 = vert_admin->n0_dof[VERTEX];
    for (i = 0; i < N_VERTICES; i++)
      fwrite(&vert_dofs[el->dof[node0 + i][n1]], sizeof(int), 1, file);
  }

  if (IS_LEAF_EL(el) || preserve_coarse_dofs)
  {
#if DIM > 1
    if (n_dof[EDGE] > 0)
    {
      node0 = node[EDGE];
      n1 = edge_admin->n0_dof[EDGE];
      for (i = 0; i < N_EDGES; i++)
	fwrite(&edge_dofs[el->dof[node0 + i][n1]], sizeof(int), 1, file);
    }
#endif

#if DIM==3
    if (n_dof[FACE] > 0)
    {
      node0 = node[FACE];
      n1 = face_admin->n0_dof[FACE];
      for (i = 0; i < N_FACES; i++)
	fwrite(&face_dofs[el->dof[node0 + i][n1]], sizeof(int), 1, file);
    }
#endif

    if ((n = n_dof[CENTER]) > 0)
    {
      node0 = node[CENTER];
      fwrite(el->dof[node0], sizeof(DOF), n, file);
    }
  }

  if (el->child[0])
  {
    for (i = 0; i < 2; i++)
      write_el_recursive(el->child[i]);
  }

  return;
}

/*--------------------------------------------------------------------------*/
/* write DOF vectors of various types                                       */
/*--------------------------------------------------------------------------*/
typedef DOF_REAL_VEC DOF_VEC;
static int write_dof_vec(const DOF_VEC *dv, const char *filename,
			 int unit_size, const char dofvectype[16])
{
  FUNCNAME("write_dof_vec");
  const FE_SPACE  *fe_space;
  const DOF_ADMIN *admin;
  MESH            *mesh;
  int             i, iadmin, last;

  if (!dv || !(fe_space = dv->fe_space))
  {
    ERROR("no %s or fe_space - no file created\n", dofvectype);
    return(1);
  }
  if (!(admin = fe_space->admin) || !(mesh = admin->mesh))
  {
    ERROR("no dof_admin or dof_admin->mesh - no file created\n");
    return(1);
  }

  dof_compress(mesh);

  iadmin = -1;
  for (i = 0; i < mesh->n_dof_admin; i++)
    if (mesh->dof_admin[i] == admin) {
      iadmin = i;
      break;
    }
  if (iadmin < 0) {
    ERROR("vec->admin not in mesh->dof_admin[] - no file created\n");
    return(1);
  }

  last = admin->used_count;
  TEST_EXIT(last <= dv->size)
    ("dof_vec->size %d < admin->size_used %d\n", dv->size, last);

  if (!(file=fopen(filename,"wb"))){
    ERROR("cannot open file %s \n",filename);
    return(1);
  }

  fwrite(dofvectype, 16, 1, file);

  i = dv->name ? strlen(dv->name) : 0;
  fwrite(&i, sizeof(int), 1, file);
  if (i) fwrite(dv->name, sizeof(char), i, file);


  fwrite(admin->n_dof, sizeof(int), (DIM+1), file);


  if (fe_space->bas_fcts) {
    i = fe_space->bas_fcts->name ? strlen(fe_space->bas_fcts->name) : 0;
    fwrite(&i, sizeof(int), 1, file);
    if (i) fwrite(fe_space->bas_fcts->name, sizeof(char), i, file);
  }
  else {
    i = 0;
    fwrite(&i, sizeof(int), 1, file);
  }


  fwrite(&last, sizeof(int), 1, file);
  fwrite(dv->vec, unit_size, last, file);

  fwrite("EOF.", 4, 1, file);                            /* file end marker */
  fclose(file);
  
  file = nil;
  return(0);
}
/*--------------------------------------------------------------------------*/

int write_dof_real_vec(const DOF_REAL_VEC *dv, const char *filename)
{
  return(write_dof_vec((const DOF_VEC *)dv, filename,
		       sizeof(REAL), "DOF_REAL_VEC    "));
}

int write_dof_real_d_vec(const DOF_REAL_D_VEC *dv, const char *filename)
{
  return(write_dof_vec((const DOF_VEC *)dv, filename,
		       sizeof(REAL_D), "DOF_REAL_D_VEC  "));
}

int write_dof_int_vec(const DOF_INT_VEC *dv, const char *filename)
{
  return(write_dof_vec((const DOF_VEC *)dv, filename,
		       sizeof(int), "DOF_INT_VEC     "));
}

int write_dof_schar_vec(const DOF_SCHAR_VEC *dv, const char *filename)
{
  return(write_dof_vec((const DOF_VEC *)dv, filename,
		       sizeof(S_CHAR), "DOF_SCHAR_VEC   "));
}

int write_dof_uchar_vec(const DOF_UCHAR_VEC *dv, const char *filename)
{
  return(write_dof_vec((const DOF_VEC *)dv, filename,
		       sizeof(U_CHAR), "DOF_UCHAR_VEC   "));
}
/*--------------------------------------------------------------------------*/
